import json
import os
import traceback

from flask import jsonify, redirect, render_template, request
from flask_login import login_required
from framework import F
from support import AlchemyEncoder


def default_route(P):
    @P.blueprint.route('/')
    def home():
        if P.ModelSetting is not None:
            tmp = P.ModelSetting.get('recent_menu_plugin')
            if tmp is not None and tmp != '':
                tmps = tmp.split('|')
                if len(tmps) == 2:
                    return redirect('/{package_name}/{sub}/{sub2}'.format(package_name=P.package_name, sub=tmps[0], sub2=tmps[1]))
                elif len(tmps) == 1 and not (P.package_name =='system' and tmps[0] == 'logout'):
                    return redirect('/{package_name}/{sub}'.format(package_name=P.package_name, sub=tmps[0]))

        return redirect('/{package_name}/{home_module}'.format(package_name=P.package_name, home_module=P.home_module))
        
    @P.blueprint.route('/<sub>', methods=['GET', 'POST'])
    @login_required
    def first_menu(sub):
        try:
            if P.ModelSetting is not None and (P.package_name == 'system' and sub != 'home'):
                P.ModelSetting.set('recent_menu_plugin', '{}'.format(sub))
            for module in P.module_list:
                if sub == module.name:
                    first_menu =  module.get_first_menu()
                    if first_menu:
                        return redirect('/{package_name}/{sub}/{first_menu}'.format(package_name=P.package_name, sub=sub, first_menu=module.get_first_menu()))
                    else:
                        return module.process_menu(None, request)
            if sub == 'log':
                return render_template('log.html', package=P.package_name)
            elif sub == 'manual':
                #return redirect(f"/{P.package_name}/manual/{P.menu['second']['manual'][0][0]}")
                try:
                    return redirect(f"/{P.package_name}/manual/{P.menu['sub2']['manual'][0][0]}")
                except:
                    return redirect(f"/{P.package_name}/manual/{P.get_first_manual_path()}")
                    
            return render_template('sample.html', title='%s - %s' % (P.package_name, sub))
        except Exception as e:
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())
    
    @P.blueprint.route('/manual/<path:path>', methods=['GET', 'POST'])
    @login_required
    def manual(path):
        try:
            plugin_root = os.path.dirname(P.blueprint.template_folder)
            filepath = os.path.join(plugin_root,  *path.split('/'))
            from support import SupportFile
            data = SupportFile.read_file(filepath)
            if filepath.endswith('.mdf'):
                try:
                    from support import SupportSC
                    data = SupportSC.decode(data)
                except:
                    pass
            return render_template('manual.html', data=data)
        except Exception as e:
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())
    
    @P.blueprint.route('/<module_name>/manual/<path:path>', methods=['GET', 'POST'])
    @login_required
    def module_manual(module_name, path):
        try:
            plugin_root = os.path.dirname(P.blueprint.template_folder)
            filepath = os.path.join(plugin_root,  *path.split('/'))
            from support import SupportFile
            data = SupportFile.read_file(filepath)
            return render_template('manual.html', data=data)
        except Exception as e:
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())


    @P.blueprint.route('/<sub>/<sub2>', methods=['GET', 'POST'])
    @login_required
    def second_menu(sub, sub2):
        if sub2 == 'null':
            return
        if P.ModelSetting is not None:
            P.ModelSetting.set('recent_menu_plugin', '{}|{}'.format(sub, sub2))
        try:
            for module in P.module_list:
                if sub == module.name:
                    return module.process_menu(sub2, request)
            if sub == 'log':
                return render_template('log.html', package=P.package_name)
            return render_template('sample.html', title='%s - %s' % (P.package_name, sub))
        except Exception as e:
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    #########################################################
    # For UI
    #########################################################
    @P.blueprint.route('/ajax/<sub>', methods=['GET', 'POST'])
    @login_required
    def ajax(sub):
        #P.logger.debug('AJAX %s %s', P.package_name, sub)
        try:
            # global
            if sub == 'setting_save':
                ret, change_list = P.ModelSetting.setting_save(request)
                for module in P.module_list:
                    module.setting_save_after(change_list)
                return jsonify(ret)
            
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())  

    @P.blueprint.route('/ajax/<module_name>/<cmd>', methods=['GET', 'POST'])
    @login_required
    def second_ajax(module_name, cmd):
        try:
            for module in P.module_list:
                if cmd == 'scheduler':
                    go = request.form['scheduler']
                    if go == 'true':
                        P.logic.scheduler_start(module_name)
                    else:
                        P.logic.scheduler_stop(module_name)
                    return jsonify(go)
                elif cmd == 'db_delete':
                    day = request.form['day']
                    ret = P.logic.db_delete(module_name, None, day)
                    return jsonify(ret)
                elif cmd == 'one_execute':
                    ret = P.logic.one_execute(module_name)
                    return jsonify(ret)
                elif cmd == 'immediately_execute':
                    ret = P.logic.immediately_execute(module_name)
                    return jsonify(ret)
                
                    
                if module_name == module.name:
                    if cmd == 'command':
                        return module.process_command(request.form['command'], request.form.get('arg1'), request.form.get('arg2'), request.form.get('arg3'), request)
                    elif cmd == 'web_list':
                        model = P.logic.get_module(module_name).web_list_model
                        if model != None:
                            return jsonify(model.web_list(request))
                    elif cmd == 'db_delete_item':
                        db_id = request.form['db_id']
                        ret = False
                        if module.web_list_model != None:
                            ret = module.web_list_model.delete_by_id(db_id)
                        return jsonify(ret)
                    else:
                        return module.process_ajax(cmd, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    @P.blueprint.route('/ajax/<module_name>/<page_name>/<command>', methods=['GET', 'POST'])
    @login_required
    def sub_ajax(module_name, page_name, command):
        try:
            ins_module = P.logic.get_module(module_name)
            ins_page = ins_module.get_page(page_name)
            if ins_page != None:
                if command == 'scheduler':
                    #sub = page_name
                    go = request.form['scheduler']
                    P.logger.debug('scheduler :%s', go)
                    if go == 'true':
                        P.logic.scheduler_start_sub(module_name, page_name)
                    else:
                        P.logic.scheduler_stop_sub(module_name, page_name)
                    return jsonify(go)
                elif command == 'db_delete':
                    day = request.form['day']
                    ret = P.logic.db_delete(module_name, page_name, day)
                    return jsonify(ret)
                elif command == 'one_execute':
                    ret = P.logic.one_execute_sub(module_name, page_name)
                    return jsonify(ret)
                elif command == 'immediately_execute':
                    ret = P.logic.immediately_execute_sub(module_name, page_name)
                    return jsonify(ret)
                elif command == 'command':
                    return ins_page.process_command(request.form['command'], request.form.get('arg1'), request.form.get('arg2'), request.form.get('arg3'), request)
                elif command == 'db_delete_item':
                    db_id = request.form['db_id']
                    ret = False
                    if ins_page.web_list_model != None:
                        ret = ins_page.web_list_model.delete_by_id(db_id)
                    return jsonify(ret)
                else:
                    return ins_page.process_ajax(command, request)
            P.logger.error(f"not process ajax : {P.package_name} {module_name} {page_name} {command}")
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    #########################################################
    # API - 외부
    #########################################################
    # 단일 모듈인 경우 모듈이름을 붙이기 불편하여 추가.
    @P.blueprint.route('/api/<sub>', methods=['GET', 'POST'])
    @F.check_api
    def api_first(sub):
        try:
            return P.module_list[0].process_api(sub, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    @P.blueprint.route('/api/<sub>/<sub2>', methods=['GET', 'POST'])
    @F.check_api
    def api(sub, sub2):
        try:
            for module in P.module_list:
                if sub == module.name:
                    return module.process_api(sub2, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    @P.blueprint.route('/normal/<sub>', methods=['GET', 'POST'])
    def normal_first(sub):
        try:
            return P.module_list[0].process_normal(sub, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    @P.blueprint.route('/normal/<sub>/<sub2>', methods=['GET', 'POST'])
    def normal(sub, sub2):
        try:
            for module in P.module_list:
                if sub == module.name:
                    return module.process_normal(sub2, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())



    # default_route 끝


























def default_route_single_module(P):
    @P.blueprint.route('/')
    def home():
        return redirect('/{package_name}/{home_module}'.format(package_name=P.package_name, home_module=P.home_module))
        
    @P.blueprint.route('/<sub>', methods=['GET', 'POST'])
    @login_required
    def first_menu(sub):
        if sub == 'log':
            return render_template('log.html', package=P.package_name)
        return P.module_list[0].process_menu(sub, request)

    @P.blueprint.route('/ajax/<sub>', methods=['GET', 'POST'])
    @login_required
    def ajax(sub):
        P.logger.debug('AJAX %s %s', P.package_name, sub)
        try:
            # global
            if sub == 'setting_save':
                ret, change_list = P.ModelSetting.setting_save(request)
                if ret:
                    P.module_list[0].setting_save_after(change_list)
                    if P.module_list[0].page_list is not None:
                        for page_ins in P.module_list[0].page_list:
                            page_ins.setting_save_after(change_list)
                return jsonify(ret)
            elif sub == 'scheduler':
                sub = request.form['sub']
                go = request.form['scheduler']
                P.logger.debug('scheduler :%s', go)
                if go == 'true':
                    P.logic.scheduler_start(sub)
                else:
                    P.logic.scheduler_stop(sub)
                return jsonify(go)
            elif sub == 'one_execute':
                sub = request.form['sub']
                ret = P.logic.one_execute(sub)
                return jsonify(ret)
            else:
                return P.module_list[0].process_ajax(sub, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())  

    @P.blueprint.route('/api/<sub>', methods=['GET', 'POST'])
    @F.check_api
    def api(sub):
        try:
            return P.module_list[0].process_api(sub, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())

    @P.blueprint.route('/normal/<sub>', methods=['GET', 'POST'])
    def normal(sub):
        try:
            return P.module_list[0].process_normal(sub, request)
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())    

























# 웹은 페이지
# 파이썬은 모듈 단위일 떄 attach 사용
#   var socket11 = io.connect(window.location.href);
def default_route_socketio_module(module, attach=''):
    P = module.P
    if module.socketio_list is None:
        module.socketio_list = []

    @F.socketio.on('connect', namespace=f'/{P.package_name}/{module.name}{attach}')
    def connect():
        try:
            P.logger.debug(f'socket_connect : {P.package_name} - {module.name}{attach}')
            module.socketio_list.append(request.sid)
            socketio_callback('start', '')
            module.socketio_connect()
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())


    @F.socketio.on('disconnect', namespace=f'/{P.package_name}/{module.name}{attach}')
    def disconnect():
        try:
            P.logger.debug(f'socket_disconnect : {P.package_name} - {module.name}{attach}')
            module.socketio_list.remove(request.sid)
            module.socketio_disconnect()
        except Exception as e: 
            P.logger.error(f"Exception:{str(e)}")
            P.logger.error(traceback.format_exc())


    def socketio_callback(cmd, data, encoding=True):
        if module.socketio_list:
            if encoding:
                data = json.dumps(data, cls=AlchemyEncoder)
                data = json.loads(data)
            F.socketio.emit(cmd, data, namespace=f'/{P.package_name}/{module.name}{attach}')

    module.socketio_callback = socketio_callback
























def default_route_socketio_page(page):
    
    module = page.parent
    P = page.P
    if page.socketio_list is None:
        page.socketio_list = []

    @F.socketio.on('connect', namespace=f'/{P.package_name}/{module.name}/{page.name}')
    def page_socketio_connect():
        try:
            #P.logger.debug(f'socket_connect : {P.package_name}/{module.name}/{page.name}')
            page.socketio_list.append(request.sid)
            page_socketio_socketio_callback('start', '')
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())


    @F.socketio.on('disconnect', namespace=f'/{P.package_name}/{module.name}/{page.name}')
    def page_socketio_disconnect():
        try:
            #P.logger.debug(f'socket_disconnect : {P.package_name}/{module.name}/{page.name}')
            page.socketio_list.remove(request.sid)
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())


    def page_socketio_socketio_callback(cmd, data, encoding=True):
        if page.socketio_list:
            if encoding:
                data = json.dumps(data, cls=AlchemyEncoder)
                data = json.loads(data)
            F.socketio.emit(cmd, data, namespace=f'/{P.package_name}/{module.name}/{page.name}')

    page.socketio_callback = page_socketio_socketio_callback



import os
import traceback

from flask import (jsonify, redirect, render_template, request,
                   send_from_directory)
from flask_login import login_required
from framework import F


@F.app.route('/global/ajax/<sub>', methods=['GET', 'POST'])
@login_required
def global_ajax(sub):
    #logger.debug('/global/ajax/%s', sub)
    if sub == 'listdir':
        if 'path' in request.form:
            #if os.path.isfile(request.form['path']):
            #    return jsonify('')
            path = request.form['path']
            if os.path.isfile(path):
                path = os.path.dirname(path)
            result_list = os.listdir(path)

            if 'only_dir' in request.form and request.form['only_dir'].lower() == 'true':
                result_list = [name for name in result_list if os.path.isdir(os.path.join(path, name))]

            result_list.sort()
            result_list = [f"{x}|{os.path.join(path,x)}" for x in result_list]
            tmp = os.path.dirname(path)
            if path != tmp:
                result_list = [f'..|{tmp}'] + result_list
            return jsonify(result_list)
        else:
            return jsonify(None)    
    elif sub == 'is_available_edit':
        # globalEditBtn
        try:
            if F.PluginManager.all_package_list['flaskcode']['loading']:
                PP = F.PluginManager.all_package_list['flaskcode']['P']
            #import flaskcode
            #from flaskcode.setup import P as PP
                ret = {'ret':True, 'target':PP.ModelSetting.get('setting_open_target')}
                return jsonify(ret)
            else:
                return jsonify({'ret':False})
            
        except:
            return jsonify({'ret':False})
    elif sub == 'command_modal_hide':
        from tool import ToolModalCommand
        ToolModalCommand.modal_close()
        return jsonify('')
    elif sub == 'command_modal_input':
        from tool import ToolModalCommand
        cmd = request.form['cmd']
        ToolModalCommand.input_command(cmd)
        return jsonify('')



@F.app.route('/robots.txt')
def robot_to_root():
    return send_from_directory('', 'static/file/robots.txt')




@F.app.route("/")
@F.app.route("/None")
@F.app.route("/home")
def home():
    return redirect('/system/home')


@F.app.route("/version")
def get_version():
    from .version import VERSION
    return VERSION

@F.app.route("/open/<path:path>")
@login_required
def open_file(path):
    return send_from_directory('/', path)


@F.app.route("/file/<path:path>")
@F.check_api
def file2(path):
    import platform
    if platform.system() == 'Windows':
        path = os.path.splitdrive(path)[1][1:]
    return send_from_directory('/', path, as_attachment=True)


@F.app.route("/upload", methods=['GET', 'POST'])
def upload():
    try:
        if request.method == 'POST':
            f = request.files['file']
            from werkzeug.utils import secure_filename
            upload_path = F.SystemModelSetting.get('path_upload')
            os.makedirs(upload_path, exist_ok=True)
            f.save(os.path.join(upload_path, secure_filename(f.filename)))
            return jsonify('success')
    except Exception as e:
        F.logger.error(f"Exception:{str(e)}")
        F.logger.error(traceback.format_exc())
        return jsonify('fail')


@F.app.route("/videojs", methods=['GET', 'POST'])
def videojs():
    data = {}
    data['play_title'] = request.form['play_title']
    data['play_source_src'] = request.form['play_source_src']
    data['play_source_type'] = request.form['play_source_type']
    if 'play_subtitle_src' in request.form:
        data['play_subtitle_src'] = request.form['play_subtitle_src']
    return render_template('videojs.html', data=data)


@F.app.route("/videojs_discord", methods=['GET', 'POST'])
def videojs_og():
    data = {}
    """
    data['play_title'] = request.form['play_title']
    data['play_source_src'] = request.form['play_source_src']
    data['play_source_type'] = request.form['play_source_type']
    if 'play_subtitle_src' in request.form:
        data['play_subtitle_src'] = request.form['play_subtitle_src']
    """
    return render_template('videojs_discord.html', data=data)


@F.app.route("/headers", methods=['GET', 'POST'])
def headers():
    from support import d
    F.logger.info(d(request.headers))
    return jsonify(d(request.headers))
    

# 3.10에서 이거 필수
@F.socketio.on('connect', namespace=f'/framework')
def connect():
    pass




import os
from functools import wraps

from flask import abort, request


def check_api(original_function):
    @wraps(original_function)
    def wrapper_function(*args, **kwargs):  #1
        from framework import F

        #logger.debug('CHECK API... {} '.format(original_function.__module__))
        #logger.warning(request.url)
        #logger.warning(request.form)
        try:
            if F.SystemModelSetting.get_bool('use_apikey'):
                try:
                    d = request.get_json()
                except Exception:
                    d = request.form.to_dict() if request.method == 'POST' else request.args.to_dict()
                apikey = d.get('apikey')
                if apikey is None or apikey != F.SystemModelSetting.get('apikey'):
                    F.logger.warning('CHECK API : ABORT no match ({})'.format(apikey))
                    F.logger.warning(request.environ.get('HTTP_X_REAL_IP', request.remote_addr))
                    abort(403)
                    return 
        except Exception as e: 
            F.logger.warning('CHECK API : ABORT exception')
            abort(403)
            return 
        return original_function(*args, **kwargs)  #2
    return wrapper_function

# Support를 logger 생성전에 쓰지 않기 위해 중복 선언 
import logging


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    green = "\x1B[32m"
    # pathname filename
    #format = "[%(asctime)s|%(name)s|%(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    
    __format = '[{yellow}%(asctime)s{reset}|{color}%(levelname)s{reset}|{green}%(name)s{reset}|%(pathname)s:%(lineno)s] {color}%(message)s{reset}' if os.environ.get('LOGGER_PATHNAME', "False") == "True" else '[{yellow}%(asctime)s{reset}|{color}%(levelname)s{reset}|{green}%(name)s{reset}|%(filename)s:%(lineno)s] {color}%(message)s{reset}'

    FORMATS = {
        logging.DEBUG: __format.format(color=grey, reset=reset, yellow=yellow, green=green),
        logging.INFO: __format.format(color=green, reset=reset, yellow=yellow, green=green),
        logging.WARNING: __format.format(color=yellow, reset=reset, yellow=yellow, green=green),
        logging.ERROR: __format.format(color=red, reset=reset, yellow=yellow, green=green),
        logging.CRITICAL: __format.format(color=bold_red, reset=reset, yellow=yellow, green=green)
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



class User:
    def __init__(self, user_id, email=None, passwd_hash=None, authenticated=False):
        self.user_id = user_id
        self.email = email
        self.passwd_hash = passwd_hash
        self.authenticated = authenticated

    def __repr__(self):
        r = {
            'user_id': self.user_id,
            'email': self.email,
            'passwd_hash': self.passwd_hash,
            'authenticated': self.authenticated,
        }
        return str(r)

    def can_login(self, passwd_hash):
        #from support import SupportAES
        #tmp = SupportAES.decrypt(self.passwd_hash)
        import hashlib
        enc = hashlib.md5()
        enc.update(passwd_hash.encode())
        hash = enc.hexdigest()
        #return True
        return self.passwd_hash == hash

    def is_active(self):
        return True

    def get_id(self):
        return self.user_id

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False





import enum
import os
import platform
import re
import shutil
import subprocess
import threading
import time
import traceback
from datetime import datetime

from support import SupportFile, SupportSubprocess, SupportUtil, logger


class SupportFfmpeg(object):
    __instance_list = []
    __ffmpeg_path = None

    __idx = 1
    total_callback_function = None
    temp_path = None

    @classmethod
    def initialize(cls, __ffmpeg_path, temp_path, total_callback_function, max_pf_count=-1):
        cls.__ffmpeg_path = __ffmpeg_path
        cls.temp_path = temp_path
        cls.total_callback_function = total_callback_function
        cls.max_pf_count = max_pf_count

    # retry : 재시도 횟수 
    # max_error_packet_count : 이 숫자 초과시 중단
    # where : 호출 모듈
    def __init__(self, url, filename, save_path=None, max_pf_count=None, headers=None, timeout_minute=60, proxy=None, callback_id=None, callback_function=None):
        self.__idx = str(SupportFfmpeg.__idx)
        SupportFfmpeg.__idx += 1

        self.url = url
        self.filename = filename
        self.save_path = save_path
        self.max_pf_count = max_pf_count
        self.headers = headers
        self.timeout_minute = int(timeout_minute)
        self.proxy = proxy
        self.callback_id = callback_id
        if callback_id == None:
            self.callback_id = str(self.__idx)
        self.callback_function = callback_function

        self.temp_fullpath = os.path.join(self.temp_path, filename)
        self.save_fullpath = os.path.join(self.save_path, filename)
        self.thread = None
        self.process = None
        self.log_thread = None
        self.status = SupportFfmpeg.Status.READY
        self.duration = 0
        self.duration_str = ''
        self.current_duration = 0
        self.percent = 0
        #self.log = []
        self.current_pf_count = 0
        self.current_bitrate = ''
        self.current_speed = ''
        self.start_time = None
        self.end_time = None
        self.download_time = None
        self.start_event = threading.Event()
        self.exist = False
        self.filesize = 0
        self.filesize_str = ''
        self.download_speed = ''
        
        
        SupportFfmpeg.__instance_list.append(self)
        if len(SupportFfmpeg.__instance_list) > 30:
            for instance in SupportFfmpeg.__instance_list:
                if instance.thread is None and instance.status != SupportFfmpeg.Status.READY:
                    SupportFfmpeg.__instance_list.remove(instance)
                    break
                else:
                    logger.debug('remove fail %s %s', instance.thread, self.status)
        
    def start(self):  
        self.thread = threading.Thread(target=self.thread_fuction, args=())
        self.thread.start()
        self.start_time = datetime.now()
        return self.get_data()
    
    def start_and_wait(self):
        self.start()
        self.thread.join(timeout=60*70)

    def stop(self):
        try:
            self.status = SupportFfmpeg.Status.USER_STOP
            self.kill()
            #logger.warning('stop')
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
    
    def kill(self):
        try:
            if self.process is not None and self.process.poll() is None:
                import psutil
                process = psutil.Process(self.process.pid)
                for proc in process.children(recursive=True):
                    proc.kill()
                process.kill()
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
            

    def thread_fuction(self):
        try:
            header_count = 0
            if self.proxy is None:
                if self.headers is None:
                    if platform.system() == 'Windows':
                        command = [self.__ffmpeg_path, '-y', '-i', f'"{self.url}"', '-c', 'copy', '-bsf:a', 'aac_adtstoasc']
                    else:
                        command = [self.__ffmpeg_path, '-y', '-i', self.url, '-c', 'copy', '-bsf:a', 'aac_adtstoasc']

                else:
                    headers_command = []
                    tmp = ""
                    for key, value in self.headers.items():
                        if key.lower() == 'user-agent':
                            headers_command.append('-user_agent')
                            headers_command.append(f"{value}")
                            pass
                        else:
                            #headers_command.append('-headers')
                            if platform.system() == 'Windows':
                                tmp += f'{key}:{value}\r\n'
                                header_count += 1
                            else:
                                #tmp.append(f'{key}:{value}')
                                tmp += f'{key}:{value}\r\n'
                    if len(tmp) > 0:
                        headers_command.append('-headers')
                        headers_command.append(f'{tmp}')
                    if platform.system() == 'Windows':
                        command = [self.__ffmpeg_path, '-y'] + headers_command + ['-i', f'"{self.url}"', '-c', 'copy', '-bsf:a', 'aac_adtstoasc']
                    else:
                        command = [self.__ffmpeg_path, '-y'] + headers_command + ['-i', self.url, '-c', 'copy', '-bsf:a', 'aac_adtstoasc']
            else:
                if platform.system() == 'Windows':
                    command = [self.__ffmpeg_path, '-y', '-http_proxy', self.proxy, '-i', f'"{self.url}"', '-c', 'copy', '-bsf:a', 'aac_adtstoasc']
                else:
                    command = [self.__ffmpeg_path, '-y', '-http_proxy', self.proxy, '-i', self.url, '-c', 'copy', '-bsf:a', 'aac_adtstoasc']

            
            if platform.system() == 'Windows':
                now = str(datetime.now()).replace(':', '').replace('-', '').replace(' ', '-')
                filename = ('%s' % now) + '.mp4'
                self.temp_fullpath = os.path.join(self.temp_path, filename)
                command.append(self.temp_fullpath)
            else:
                command.append(self.temp_fullpath)
            

            try:
                #logger.debug(' '.join(command))
                if os.path.exists(self.temp_fullpath):
                    for f in SupportFfmpeg.__instance_list:
                        if f.__idx != self.__idx and f.temp_fullpath == self.temp_fullpath and f.status in [SupportFfmpeg.Status.DOWNLOADING, SupportFfmpeg.Status.READY]:
                            self.status = SupportFfmpeg.Status.ALREADY_DOWNLOADING
                            return
            except:
                pass
            #logger.error(' '.join(command))
            command = SupportSubprocess.command_for_windows(command)

            if platform.system() == 'Windows' and header_count > 1:
                if os.environ.get('FF'):
                    from framework import F
                    batfilepath = os.path.join(F.config['path_data'], 'tmp', f"{time.time()}.bat")
                else:
                    batfilepath = f"{time.time()}.bat"
                tmp = command.replace('\r\n', '!CRLF!')
                text = f"""setlocal enabledelayedexpansion
SET CRLF=^


{tmp}"""
                SupportFile.write_file(batfilepath, text)
                command = batfilepath
            
            self.process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, encoding='utf8')

            self.status = SupportFfmpeg.Status.READY
            self.log_thread = threading.Thread(target=self.log_thread_fuction, args=())
            self.log_thread.start()
            self.start_event.wait(timeout=60)
            if self.log_thread is None:
                if self.status == SupportFfmpeg.Status.READY:
                    self.status = SupportFfmpeg.Status.ERROR
                self.kill()
            elif self.status == SupportFfmpeg.Status.READY:
                self.status = SupportFfmpeg.Status.ERROR
                self.kill()
            else:
                process_ret = self.process.wait(timeout=60*self.timeout_minute)
                # 2022-10-25
                time.sleep(3)
                logger.info(f"{process_ret=}")
                if process_ret is None: # timeout
                    if self.status != SupportFfmpeg.Status.COMPLETED and self.status != SupportFfmpeg.Status.USER_STOP and self.status != SupportFfmpeg.Status.PF_STOP:
                        self.status = SupportFfmpeg.Status.TIME_OVER
                        self.kill()
                else:
                    if self.status == SupportFfmpeg.Status.DOWNLOADING:
                        self.status = SupportFfmpeg.Status.FORCE_STOP
            self.end_time = datetime.now()
            self.download_time = self.end_time - self.start_time
            try:
                if self.status == SupportFfmpeg.Status.COMPLETED:
                    if self.save_fullpath != self.temp_fullpath:
                        if os.path.exists(self.save_fullpath):
                            os.remove(self.save_fullpath)
                        if platform.system() != 'Windows':
                            os.system('chmod 777 "%s"' % self.temp_fullpath)
                        shutil.move(self.temp_fullpath, self.save_fullpath)
                        self.filesize = os.stat(self.save_fullpath).st_size
                else:
                    if os.path.exists(self.temp_fullpath):
                        os.remove(self.temp_fullpath)
            except Exception as e:
                logger.error(f"Exception:{str(e)}")
                logger.error(traceback.format_exc())

            arg = {'type':'last', 'status':self.status, 'data' : self.get_data()}
            self.send_to_listener(**arg)
            self.process = None
            self.thread = None
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
            try:
                self.status = SupportFfmpeg.Status.EXCEPTION
                arg = {'type':'last', 'status':self.status, 'data' : self.get_data()}
                self.send_to_listener(**arg)
            except:
                pass

            

    def log_thread_fuction(self):
        with self.process.stdout:
            for line in iter(self.process.stdout.readline, ''):
                line = line.strip()
                #logger.error(line)
                try:
                    if self.status == SupportFfmpeg.Status.READY:
                        if line.find('Server returned 404 Not Found') != -1 or line.find('Unknown error') != -1:
                            self.status = SupportFfmpeg.Status.WRONG_URL
                            self.start_event.set()
                        elif line.find('No such file or directory') != -1:
                            self.status = SupportFfmpeg.Status.WRONG_DIRECTORY
                            self.start_event.set()
                        else:
                            match = re.compile(r'Duration\:\s(\d{2})\:(\d{2})\:(\d{2})\.(\d{2})\,\sstart').search(line)
                            if match:
                                self.duration_str = '%s:%s:%s' % ( match.group(1), match.group(2), match.group(3))
                                self.duration = int(match.group(4))
                                self.duration += int(match.group(3)) * 100
                                self.duration += int(match.group(2)) * 100 * 60
                                self.duration += int(match.group(1)) * 100 * 60 * 60
                                if match:
                                    self.status = SupportFfmpeg.Status.READY
                                    arg = {'type':'status_change', 'status':self.status, 'data' : self.get_data()}
                                    self.send_to_listener(**arg)
                                continue
                            match = re.compile(r'time\=(\d{2})\:(\d{2})\:(\d{2})\.(\d{2})\sbitrate\=\s*(?P<bitrate>\d+).*?[$|\s](\s?speed\=\s*(?P<speed>.*?)x)?').search(line)
                            if match:
                                self.status = SupportFfmpeg.Status.DOWNLOADING
                                arg = {'type':'status_change', 'status':self.status, 'data' : self.get_data()}
                                self.send_to_listener(**arg)
                                self.start_event.set()
                    elif self.status == SupportFfmpeg.Status.DOWNLOADING:
                        if line.find('PES packet size mismatch') != -1:
                            self.current_pf_count += 1
                            if self.current_pf_count > self.max_pf_count:
                                self.status = SupportFfmpeg.Status.PF_STOP
                                self.kill()
                            continue
                        if line.find('HTTP error 403 Forbidden') != -1:
                            self.status = SupportFfmpeg.Status.HTTP_FORBIDDEN
                            self.kill()
                            continue
                        match = re.compile(r'time\=(\d{2})\:(\d{2})\:(\d{2})\.(\d{2})\sbitrate\=\s*(?P<bitrate>\d+).*?[$|\s](\s?speed\=\s*(?P<speed>.*?)x)?').search(line)
                        if match: 
                            self.current_duration = int(match.group(4))
                            self.current_duration += int(match.group(3)) * 100
                            self.current_duration += int(match.group(2)) * 100 * 60
                            self.current_duration += int(match.group(1)) * 100 * 60 * 60
                            try:
                                self.percent = int(self.current_duration * 100 / self.duration)
                            except: pass
                            self.current_bitrate = match.group('bitrate')
                            self.current_speed = match.group('speed')
                            self.download_time = datetime.now() - self.start_time
                            arg = {'type':'normal', 'status':self.status, 'data' : self.get_data()}
                            self.send_to_listener(**arg)
                            continue
                        match = re.compile(r'video\:\d*kB\saudio\:\d*kB').search(line)
                        if match:
                            self.status = SupportFfmpeg.Status.COMPLETED
                            self.end_time = datetime.now()
                            self.download_time = self.end_time - self.start_time
                            self.percent = 100
                            arg = {'type':'status_change', 'status':self.status, 'data' : self.get_data()}
                            self.send_to_listener(**arg)
                            continue
                except Exception as e:
                    logger.error(f'Exception:{str(e)}')
                    logger.error(traceback.format_exc())
        self.start_event.set()
        self.log_thread = None


    def get_data(self):
        data = {
            'url' : self.url,
            'filename' : self.filename,
            'max_pf_count' : self.max_pf_count,
            'callback_id' : self.callback_id,
            'temp_path' : self.temp_path,
            'save_path' : self.save_path,
            'temp_fullpath' : self.temp_fullpath,
            'save_fullpath' : self.save_fullpath,
            'status' : int(self.status),
            'status_str' : self.status.name,
            'status_kor' : str(self.status),
            'duration' : self.duration,
            'duration_str' : self.duration_str,
            'current_duration' : self.current_duration,
            'percent' : self.percent,
            'current_pf_count' : self.current_pf_count,
            'idx' : self.__idx,
            #'log' : self.log,
            'current_bitrate' : self.current_bitrate,
            'current_speed' : self.current_speed,
            'start_time' : '' if self.start_time is None else str(self.start_time).split('.')[0][5:],
            'end_time' : '' if self.end_time is None else str(self.end_time).split('.')[0][5:],
            'download_time' : '' if self.download_time is None else '%02d:%02d' % (self.download_time.seconds/60, self.download_time.seconds%60),
            'exist' : os.path.exists(self.save_fullpath),
        }                        
        if self.status == SupportFfmpeg.Status.COMPLETED:
            data['filesize'] = self.filesize
            data['filesize_str'] = SupportUtil.sizeof_fmt(self.filesize)
            if self.download_time.seconds != 0:
                data['download_speed'] = SupportUtil.sizeof_fmt(self.filesize/self.download_time.seconds, suffix='Bytes/Second')
            else:
                data['download_speed'] = '0Bytes/Second'
        return data

    def send_to_listener(self, **arg):
        if self.total_callback_function != None:
            self.total_callback_function(**arg)
        if self.callback_function is not None and self.callback_function != self.total_callback_function:
            arg['callback_id'] = self.callback_id
            self.callback_function(**arg)          


    @classmethod
    def stop_by_idx(cls, idx):
        try:
            for __instance in SupportFfmpeg.__instance_list:
                if __instance.__idx == idx:
                    __instance.stop()
                    break
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
    

    @classmethod
    def stop_by_callback_id(cls, callback_id):
        try:
            for __instance in SupportFfmpeg.__instance_list:
                if __instance.callback_id == callback_id:
                    __instance.stop()
                    break
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())


    @classmethod
    def get_instance_by_idx(cls, idx):
        try:
            for __instance in SupportFfmpeg.__instance_list:
                if __instance.__idx == idx:
                    return __instance
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())

    

    @classmethod
    def get_instance_by_callback_id(cls, callback_id):
        try:
            for __instance in SupportFfmpeg.__instance_list:
                if __instance.callback_id == callback_id:
                    return __instance
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
        
    @classmethod
    def all_stop(cls):
        try:
            for __instance in SupportFfmpeg.__instance_list:
                __instance.stop()
        except Exception as e:
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())

    @classmethod
    def get_list(cls):
        return cls.__instance_list






    class Status(enum.Enum):
        READY = 0
        WRONG_URL = 1
        WRONG_DIRECTORY = 2
        EXCEPTION = 3
        ERROR = 4
        HTTP_FORBIDDEN = 11
        
        DOWNLOADING = 5

        USER_STOP = 6
        COMPLETED = 7
        TIME_OVER = 8
        PF_STOP = 9
        FORCE_STOP = 10 #강제중단
        ALREADY_DOWNLOADING = 12 #이미 목록에 있고 다운로드중

        def __int__(self):
            return self.value

        def __str__(self):
            kor = ['준비', 'URL에러', '폴더에러', '실패(Exception)', '실패(에러)', '다운로드중', '사용자중지', '완료', '시간초과', 'PF중지', '강제중지',
            '403에러', '임시파일이 이미 있음']
            return kor[int(self)]
        
        def __repr__(self):
            return self.name
        
        @staticmethod
        def get_instance(value):
            tmp = [ 
                SupportFfmpeg.Status.READY, 
                SupportFfmpeg.Status.WRONG_URL, 
                SupportFfmpeg.Status.WRONG_DIRECTORY, 
                SupportFfmpeg.Status.EXCEPTION, 
                SupportFfmpeg.Status.ERROR,
                SupportFfmpeg.Status.DOWNLOADING, 
                SupportFfmpeg.Status.USER_STOP, 
                SupportFfmpeg.Status.COMPLETED, 
                SupportFfmpeg.Status.TIME_OVER, 
                SupportFfmpeg.Status.PF_STOP, 
                SupportFfmpeg.Status.FORCE_STOP, 
                SupportFfmpeg.Status.HTTP_FORBIDDEN, 
                SupportFfmpeg.Status.ALREADY_DOWNLOADING ]
            return tmp[value]





try:
    from gevent import monkey;monkey.patch_all()
    print('[MAIN] gevent mokey patch!!')
except:
    print('[MAIN] gevent not installed!!')
import os
import sys
import traceback

try:
    import platform
    cmd = 'export'
    if platform.system() == 'Windows':
        cmd = 'set'
    os.system(f"{cmd} CELERYD_HIJACK_ROOT_LOGGER=false")
except:pass

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'lib'))

try:
    # 2024.06.13 
    from framework import initiaize
    frame = initiaize()
    
    # flaskfarm.main : 패키지로 실행. 패키지로 celry 실행 체크
    if __name__ in ['__main__', 'flaskfarm.main'] and sys.argv[0].endswith('celery') == False:
        frame.start()
    else:  
        app = frame.app
        celery = frame.celery
except Exception as exception:
    print(str(exception))  
    print(traceback.format_exc())



import os
import traceback

from flask import Blueprint
from framework import F
from support import SupportYaml

from . import (Logic, default_route, default_route_single_module,
               get_model_setting)


class PluginBase(object):
    package_name = None
    logger = None
    blueprint = None
    menu = None
    plugin_info = None
    ModelSetting = None
    logic = None
    module_list = None
    home_module = None
    vars = []

    def __init__(self, setting):
        try:
            is_system = ('system' == os.path.basename(os.path.dirname(setting['filepath'])))
            self.status = ""
            self.setting = setting

            info_filepath = os.path.join(os.path.dirname(setting['filepath']), 'info.yaml')
            if os.path.exists(info_filepath) == False and is_system == False:
                return
            if is_system:
                self.package_name = 'system'
            else:
                self.plugin_info = SupportYaml.read_yaml(info_filepath)
                self.package_name = self.plugin_info['package_name']

            self.logger = F.get_logger(self.package_name)
            self.blueprint = Blueprint(self.package_name, self.package_name, url_prefix=f'/{self.package_name}', template_folder=os.path.join(os.path.dirname(setting['filepath']), 'templates'), static_folder=os.path.join(os.path.dirname(setting['filepath']), 'static'))
            self.menu = setting['menu']
            self.setting_menu = setting.get('setting_menu', None)

            self.ModelSetting = None
            if setting.get('use_db', True):
                db_path = os.path.join(F.config['path_data'], 'db', f'{self.package_name}.db')
                F.app.config['SQLALCHEMY_BINDS'][self.package_name] = f"sqlite:///{db_path}?check_same_thread=False"
            if setting.get('use_default_setting', True):
                self.ModelSetting = get_model_setting(self.package_name, self.logger)
            
            self.module_list = []
            self.home_module = setting.get('home_module')
            self.status = "init_success"
            self.config = {}
        except Exception as e: 
            self.logger.error(f'Exception:{str(e)}')
            self.logger.error(traceback.format_exc())
            self.status = 'init_fail'

    def set_module_list(self, mod_list):
        try:
            for mod in mod_list:
                mod_ins = mod(self)
                self.module_list.append(mod_ins)
            if self.home_module == None:
                self.home_module = self.module_list[0].name
        except Exception as e:
            F.logger.error(f'[{self.package_name}] Exception:{str(e)}')
            F.logger.error(traceback.format_exc())

            
        self.logic = Logic(self)
        route_mode = self.setting.get('default_route', 'normal')
        if route_mode == 'normal':
            default_route(self)
        elif route_mode == 'single':
            default_route_single_module(self)


    def plugin_load(self):
        if self.logic:
            self.logic.plugin_load()
    
    def plugin_load_celery(self):
        if self.logic:
            self.logic.plugin_load_celery()

    
    def plugin_unload(self):
        if self.logic:
            self.logic.plugin_unload()

    def get_first_manual_path(self):
        for __ in self.menu['list']:
            if __['uri'] == 'manual' and len(__['list']) > 0:
                return __['list'][0]['uri']


def create_plugin_instance(config):
    ins = PluginBase(config)
    return ins



import traceback

from flask import render_template


class PluginModuleBase(object):
    db_default = None
 
    def __init__(self, P, first_menu=None, name=None, scheduler_desc=None):
        self.P = P
        self.scheduler_desc = scheduler_desc
        self.first_menu = first_menu
        self.name = name
        self.socketio_list = None
        self.page_list = None
        self.web_list_model = None
    
    def get_module(self, module_name):
        try:
            for module in self.P.module_list:
                if module.name == module_name:
                    return module
        except Exception as e:
            self.P.logger.error(f'Exception:{str(e)}')
            self.P.logger.error(traceback.format_exc())


    # set_module_list 대응
    def set_page_list(self, page_list):
        try:
            self.page_list = []
            for mod in page_list:
                mod_ins = mod(self.P, self)
                self.page_list.append(mod_ins)
            
        except Exception as e:
            self.P.logger.error(f'Exception:{str(e)}')
            self.P.logger.error(traceback.format_exc())
    
    def get_page(self, page_name):
        try:
            if self.page_list == None:
                return
            for page in self.page_list:
                if page_name == page.name:
                    return page
        except Exception as e:
            self.P.logger.error(f'Exception:{str(e)}')
            self.P.logger.error(traceback.format_exc())


    def process_menu(self, page, req):
        from framework import F
        
        try:
            if self.page_list is not None:
                page_ins = self.get_page(page)
                if page_ins != None:
                    return page_ins.process_menu(req)

            arg = self.P.ModelSetting.to_dict() if self.P.ModelSetting != None else {}
            arg['path_data'] = F.config['path_data']
            arg['is_include'] = F.scheduler.is_include(self.get_scheduler_name())
            arg['is_running'] = F.scheduler.is_running(self.get_scheduler_name())
            return render_template(f'{self.P.package_name}_{self.name}_{page}.html', arg=arg)
        except Exception as e:
            self.P.logger.error(f'Exception:{str(e)}')
            self.P.logger.error(traceback.format_exc())
        return render_template('sample.html', title=f"PluginModuleBase-process_menu{self.P.package_name}/{self.name}/{page}")

    def process_ajax(self, sub, req):
        pass

    def process_command(self, command, arg1, arg2, arg3, req):
        pass

    def process_api(self, sub, req):
        pass

    def process_normal(self, sub, req):
        pass

    def scheduler_function(self):
        pass

    def db_delete(self, day):
        if self.web_list_model != None:
            return self.web_list_model.delete_all(day)


    def plugin_load(self):
        pass
    
    def plugin_load_celery(self):
        pass


    def plugin_unload(self):
        pass
    
    def setting_save_after(self, change_list):
        pass

    def process_telegram_data(self, data, target=None):
        pass

    def migration(self):
        pass
    
    #################################################################
    def get_scheduler_desc(self):
        return self.scheduler_desc 

    def get_scheduler_interval(self):
        if self.P is not None and self.P.ModelSetting is not None and self.name is not None:
            return self.P.ModelSetting.get('{module_name}_interval'.format(module_name=self.name))

    def get_first_menu(self):
        return self.first_menu

    def get_scheduler_id(self):
        return '%s_%s' % (self.P.package_name, self.name)

    def dump(self, data):
        if type(data) in [type({}), type([])]:
            import json
            return '\n' + json.dumps(data, indent=4, ensure_ascii=False)
        else:
            return str(data)

    def socketio_connect(self):
        pass

    def socketio_disconnect(self):
        pass


    def arg_to_dict(self, arg):
        return self.P.logic.arg_to_dict(arg)

    def get_scheduler_name(self):
        return f'{self.P.package_name}_{self.name}'


    def process_discord_data(self, data):
        pass

    
    def start_celery(self, func, on_message=None, *args, page=None):
        from framework import F
        if F.config['use_celery']:
            result = func.apply_async(args)
            try:
                if on_message != None:
                    ret = result.get(on_message=on_message, propagate=True)
                else:
                    ret = result.get()
            except:
                ret = result.get()
        else:
            if on_message == None:
                ret = func(*args)
            else:
                if page == None:
                    ret = func(self, *args)
                else:
                    ret = func(page, *args)
        return ret














class PluginPageBase(object):
    db_default = None

    def __init__(self, P, parent, name=None, scheduler_desc=None):
        self.P = P
        self.parent = parent
        self.name = name
        self.scheduler_desc = scheduler_desc
        self.socketio_list = None
        self.web_list_model = None

    def process_menu(self, req):
        try:
            arg = {}
            if self.P.ModelSetting != None:
                arg = self.P.ModelSetting.to_dict()
            return render_template(f'{self.P.package_name}_{self.parent.name}_{self.name}.html', arg=arg)
        except Exception as e:
            self.P.logger.error(f'Exception:{str(e)}')
            self.P.logger.error(traceback.format_exc())
            
        return render_template('sample.html', title=f"PluginPageBase-process_menu --- {self.P.package_name}/{self.parent.name}/{self.name}")


    def process_ajax(self, sub, req):
        pass
    
    def process_api(self, sub, req):
        pass

    def process_normal(self, sub, req):
        pass
    
    def process_command(self, command, arg1, arg2, arg3, req):
        pass
    
    
    
    # logic
    def plugin_load(self):
        pass
    
    def plugin_load_celery(self):
        pass
    
    # logic
    def plugin_unload(self):
        pass

    def scheduler_function(self):
        pass

    def get_scheduler_desc(self):
        return self.scheduler_desc 
    
    def get_scheduler_interval(self):
        if self.P is not None and self.P.ModelSetting is not None and self.parent.name is not None and self.name is not None:
            return self.P.ModelSetting.get(f'{self.parent.name}_{self.name}_interval')

    def get_scheduler_name(self):
        return f'{self.P.package_name}_{self.parent.name}_{self.name}'


    # logic
    def migration(self):
        pass

    
    # route
    def setting_save_after(self, change_list):
        pass

    def process_telegram_data(self, data, target=None):
        pass


    def arg_to_dict(self, arg):
        return self.P.logic.arg_to_dict(arg)


    def get_page(self, page_name):
        return self.parent.get_page(page_name)


    def get_module(self, module_name):
        return self.parent.get_module(module_name)

    def process_discord_data(self, data):
        pass
    
    def db_delete(self, day):
        if self.web_list_model != None:
            return self.web_list_model.delete_all(day)
    

    def start_celery(self, func, on_message, *args):
        return self.parent.start_celery(func, on_message, *args, page=self)
    
    """
    def start_celery(self, func, on_message=None, *args):
        from framework import F
        if F.config['use_celery']:
            result = func.apply_async(args)
            try:
                if on_message != None:
                    ret = result.get(on_message=on_message, propagate=True)
                else:
                    ret = result.get()
            except:
                ret = result.get()
        else:
            ret = func(*args)
        return ret
    """


import base64
import os
import traceback

from Crypto import Random
from Crypto.Cipher import AES

from . import logger

BS = 16
pad = lambda s: s + (BS - len(s.encode('utf-8')) % BS) * chr(BS - len(s.encode('utf-8')) % BS) 
#unpad = lambda s : s[0:-s[-1]]
unpad = lambda s : s[:-ord(s[len(s)-1:])]
key = b'140b41b22a29beb4061bda66b6747e14'

class SupportAES(object):
    @classmethod
    def encrypt(cls, raw, mykey=None):
        try:
            Random.atfork()
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 

        raw = pad(raw)
        if type(raw) == type(''):
            raw = raw.encode()
        if mykey is not None and type(mykey) == type(''):
            mykey = mykey.encode()
        iv = Random.new().read( AES.block_size )
        cipher = AES.new(key if mykey is None else mykey, AES.MODE_CBC, iv )
        try:
            tmp = cipher.encrypt( raw )
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 
            tmp = cipher.encrypt( raw.encode() )
        ret = base64.b64encode( iv + tmp ) 
        ret = ret.decode()
        return ret

    @classmethod
    def decrypt(cls, enc, mykey=None):
        enc = base64.b64decode(enc)
        iv = enc[:16]
        if len(iv) != 16:
            iv = os.urandom(16)
        if mykey is not None and type(mykey) == type(''):
            mykey = mykey.encode()
        cipher = AES.new(key if mykey is None else mykey, AES.MODE_CBC, iv )
        return unpad(cipher.decrypt( enc[16:] )).decode()

    @classmethod
    def md5(cls, text):
        import hashlib
        enc = hashlib.md5()
        enc.update(text.encode())
        return enc.hexdigest()
    



    @classmethod
    def encrypt_(cls, raw, mykey=None, iv=None):
        try:
            Random.atfork()
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 
        raw = pad(raw)
        if type(raw) == type(''):
            raw = raw.encode()
        if mykey is not None and type(mykey) == type(''):
            mykey = mykey.encode()
        if iv == None:
            iv = Random.new().read( AES.block_size )
        elif iv is not None and type(iv) == type(''):
            iv = iv.encode()
        cipher = AES.new(key if mykey is None else mykey, AES.MODE_CBC, iv )
        try:
            tmp = cipher.encrypt( raw )
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 
            tmp = cipher.encrypt( raw.encode() )
        ret = base64.b64encode( tmp ) 
        ret = ret.decode()
        return ret

    @classmethod
    def decrypt_(cls, enc, mykey=None, iv=None):
        enc = base64.b64decode(enc)
        if iv == None:
            iv = os.urandom(16)
        elif iv is not None and type(iv) == type(''):
            iv = iv.encode()
        if mykey is not None and type(mykey) == type(''):
            mykey = mykey.encode()
        cipher = AES.new(key if mykey is None else mykey, AES.MODE_CBC, iv )
        return unpad(cipher.decrypt( enc )).decode()



import re

from framework import F


def get_menu(full_query):
    match = re.compile(r'\/(?P<package_name>.*?)\/(?P<module_name>.*?)\/manual\/(?P<sub2>.*?)($|\?)').match(full_query)
    if match:
        return match.group('package_name'), match.group('module_name'), f"manual/{match.group('sub2')}"

    match = re.compile(r'\/(?P<menu>.*?)\/manual\/(?P<sub2>.*?)($|\?)').match(full_query)
    if match:
        return match.group('menu'), 'manual', match.group('sub2')

    match = re.compile(r'\/(?P<menu>.*?)\/(?P<sub>.*?)\/(?P<sub2>.*?)($|\/|\?)').match(full_query)
    if match:
        return match.group('menu'), match.group('sub'), match.group('sub2')

    match = re.compile(r'\/(?P<menu>.*?)\/(?P<sub>.*?)($|\/|\?)').match(full_query)
    if match: 
        return match.group('menu'), match.group('sub'), None

    match = re.compile(r'\/(?P<menu>.*?)($|\/|\?)').match(full_query)
    if match:
        return match.group('menu'), None , None
    return 'home', None, None
 


def get_theme():
    return F.SystemModelSetting.get('theme')

#def get_login_status():
#    if current_user is None:
#        return False
#    return current_user.is_authenticated

def get_web_title():
    try:
        return F.SystemModelSetting.get('web_title')
    except:
        return 'Home'


def is_https():
    return (F.SystemModelSetting.get('ddns').find('https://') != -1)


def jinja_initialize(app):
    #from .init_menu import get_menu_map, get_plugin_menu
    from .init_menu import MenuManager
    app.jinja_env.globals.update(get_menu=get_menu)
    app.jinja_env.globals.update(get_theme=get_theme)
    app.jinja_env.globals.update(get_menu_map=MenuManager.get_menu_map)
    app.jinja_env.globals.update(get_setting_menu=MenuManager.get_setting_menu)
    app.jinja_env.globals.update(get_web_title=get_web_title)
    app.jinja_env.globals.update(dropzone=F.dropzone)

    app.jinja_env.filters['get_menu'] = get_menu
    app.jinja_env.filters['get_theme'] = get_theme
    app.jinja_env.filters['get_menu_map'] = MenuManager.get_menu_map
    app.jinja_env.filters['get_setting_menu'] = MenuManager.get_setting_menu
    app.jinja_env.filters['get_web_title'] = get_web_title

    app.jinja_env.auto_reload = True
    app.jinja_env.add_extension('jinja2.ext.loopcontrols')



import traceback
from datetime import datetime

from framework import F


def get_model_setting(package_name, logger, table_name=None):

    class ModelSetting(F.db.Model):
        __tablename__ = '%s_setting' % package_name if table_name is None else table_name
        __table_args__ = {'mysql_collate': 'utf8_general_ci'}
        __bind_key__ = package_name

        id = F.db.Column(F.db.Integer, primary_key=True)
        key = F.db.Column(F.db.String, unique=True, nullable=False)
        value = F.db.Column(F.db.String, nullable=False)
    
        def __init__(self, key, value):
            self.key = key
            self.value = value

        def __repr__(self):
            return repr(self.as_dict())

        def as_dict(self):
            return {x.name: getattr(self, x.name) for x in self.__table__.columns}

        @staticmethod
        def get(key):
            try:
                with F.app.app_context():
                    ret = F.db.session.query(ModelSetting).filter_by(key=key).first()
                    if ret is not None:
                        return ret.value.strip()
                    return None
            except Exception as e:
                logger.error(f"Exception:{str(e)} [{key}]")
                logger.error(traceback.format_exc())

        @staticmethod
        def has_key(key):
            with F.app.app_context():
                return (F.db.session.query(ModelSetting).filter_by(key=key).first() is not None)
        
        @staticmethod
        def get_int(key):
            try:
                return int(ModelSetting.get(key))
            except Exception as e:
                logger.error(f"Exception:{str(e)} [{key}]")
                logger.error(traceback.format_exc())
        
        @staticmethod
        def get_bool(key):
            try:
                return (ModelSetting.get(key) == 'True')
            except Exception as e:
                logger.error(f"Exception:{str(e)} [{key}]")
                logger.error(traceback.format_exc())
        
        @staticmethod
        def get_datetime(key):
            try:
                tmp = ModelSetting.get(key)
                if tmp != None and tmp != '':
                    return datetime.strptime(tmp, '%Y-%m-%d %H:%M:%S.%f')
            except Exception as e:
                logger.error(f"Exception:{str(e)} [{key}]")
                logger.error(traceback.format_exc())

        @staticmethod
        def set(key, value):
            try:
                with F.app.app_context():
                    item = F.db.session.query(ModelSetting).filter_by(key=key).with_for_update().first()
                    if item is not None:
                        item.value = value.strip() if value is not None else value
                        F.db.session.commit()
                    else:
                        F.db.session.add(ModelSetting(key, value.strip()))
                        F.db.session.commit()
            except Exception as e:
                logger.error(f"Exception:{str(e)} [{key}]")
                logger.error(traceback.format_exc())

        @staticmethod
        def to_dict():
            try:
                ret = ModelSetting.db_list_to_dict(F.db.session.query(ModelSetting).all())
                ret['package_name'] = package_name
                return ret 
            except Exception as e:
                logger.error(f"Exception:{str(e)}")
                logger.error(traceback.format_exc())


        @staticmethod
        def setting_save(req):
            try:
                change_list = []
                for key, value in req.form.items():
                    if key in ['scheduler', 'is_running']:
                        continue
                    if key.startswith('global') or key.startswith('tmp_') or key.startswith('_'):
                        continue
                    #logger.debug('Key:%s Value:%s', key, value)
                    if ModelSetting.get(key) != value:
                        change_list.append(key)
                        entity = F.db.session.query(ModelSetting).filter_by(key=key).with_for_update().first()
                        entity.value = value
                F.db.session.commit()
                return True, change_list 
            except Exception as e:
                logger.error(f"Exception:{str(e)}")
                logger.error(traceback.format_exc())
                logger.debug('Error Key:%s Value:%s', key, value)
                return False, []

        @staticmethod
        def get_list(key, delimeter='\n', comment='#'):
            value = None
            try:
                value = ModelSetting.get(key).replace('\n', delimeter)
                if comment is None:
                    values = [x.strip() for x in value.split(delimeter)]
                else:
                    values = [x.split(comment)[0].strip() for x in value.split(delimeter)]
                values = ModelSetting.get_list_except_empty(values)
                return values
            except Exception as e:
                logger.error(f"Exception:{str(e)}")
                logger.error(traceback.format_exc())
                logger.error('Error Key:%s Value:%s', key, value)


        @staticmethod
        def get_list_except_empty(source):
            tmp = []
            for _ in source:
                if _.strip().startswith('#'):
                    continue
                if _.strip() != '':
                    tmp.append(_.strip())
            return tmp

        @staticmethod
        def db_list_to_dict(db_list):
            ret = {}
            for item in db_list:
                ret[item.key] = item.value
            return ret

    return ModelSetting



import io
import json
import locale
import os
import platform
import queue
import subprocess
import threading
import time
import traceback

from . import logger


def demote(user_uid, user_gid):
    def result():
        os.setgid(user_gid)
        os.setuid(user_uid)
    return result

class SupportSubprocess(object):

    @classmethod 
    def command_for_windows(cls, command: list): 
        if platform.system() == 'Windows':
            tmp = []
            if type(command) == type([]):
                for x in command:
                    if x.find(' ') == -1:
                        tmp.append(x)
                    else:
                        tmp.append(f'"{x}"')
                command = ' '.join(tmp)
        return command

    # 2021-10-25
    # timeout 적용
    @classmethod
    def execute_command_return(cls, command, format=None, log=False, shell=False, env=None, timeout=None, uid=None, gid=None):

        try:
            logger.debug(f"execute_command_return : {' '.join(command)}")
            command = cls.command_for_windows(command)

            iter_arg =  ''
            if platform.system() == 'Windows':
                process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=shell, env=env, encoding='utf8', bufsize=0)
            else:
                if uid == None:
                    process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=shell, env=env, encoding='utf8')
                else:
                    process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=shell, env=env, preexec_fn=demote(uid, gid), encoding='utf8')
            new_ret = {'status':'finish', 'log':None}

            def func(ret):
                with process.stdout:
                    try:
                        for line in iter(process.stdout.readline, iter_arg):
                            ret.append(line.strip())
                            if log:
                                logger.debug(ret[-1])
                    except:
                        pass
            
            result = []
            thread = threading.Thread(target=func, args=(result,))
            thread.setDaemon(True)
            thread.start()
            #thread.join()

            try:
                #process.communicate()
                process_ret = process.wait(timeout=timeout) # wait for the subprocess to exit
            except:
                import psutil
                process = psutil.Process(process.pid)
                for proc in process.children(recursive=True):
                    proc.kill()
                process.kill()
                new_ret['status'] = "timeout"
            #logger.error(process_ret)
            thread.join()
            #ret = []
            #with process.stdout:
            #    for line in iter(process.stdout.readline, iter_arg):
            #        ret.append(line.strip())
            #        if log:
            #            logger.debug(ret[-1])
           
            ret = result
            #logger.error(ret)
            if format is None:
                ret2 = '\n'.join(ret)
            elif format == 'json':
                try:
                    index = 0
                    for idx, tmp in enumerate(ret):
                        #logger.debug(tmp)
                        if tmp.startswith('{') or tmp.startswith('['):
                            index = idx
                            break
                    ret2 = json.loads(''.join(ret[index:]))
                except:
                    ret2 = ret

            new_ret['log'] = ret2
            return new_ret
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())
            logger.error('command : %s', command)
    

    __instance_list = []


    def __init__(self, command,  print_log=False, shell=False, env=None, timeout=None, uid=None, gid=None, stdout_callback=None, call_id=None, callback_line=True):
        self.command = command
        self.print_log = print_log
        self.shell = shell
        self.env = env
        self.timeout = timeout
        self.uid = uid
        self.gid = gid
        self.stdout_callback = stdout_callback
        self.process = None
        self.stdout_queue = None
        self.call_id = call_id
        self.timestamp = time.time()
        self.callback_line = callback_line
        

    def start(self, join=True):
        try:
            self.thread = threading.Thread(target=self.__execute_thread_function, args=())
            self.thread.setDaemon(True)
            self.thread.start()
            if join:
                self.thread.join()
        except Exception as e: 
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())


    def __execute_thread_function(self):
        try:
            self.command = self.command_for_windows(self.command)
            logger.debug(f"{self.command=}")
            if platform.system() == 'Windows':
                self.process = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=self.shell, env=self.env, encoding='utf8', bufsize=0)

            else:
                if self.uid == None:
                    self.process = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=self.shell, env=self.env, encoding='utf8', bufsize=0)
                else:
                    self.process = subprocess.Popen(self.command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, shell=self.shell, env=self.env, preexec_fn=demote(self.uid, self.gid), encoding='utf8', bufsize=0)
            SupportSubprocess.__instance_list.append(self)
            self.send_stdout_callback(self.call_id, 'START', None)
            self.__start_communicate()
            self.__start_send_callback()
            if self.process is not None:
                if self.timeout != None:
                    self.process.wait(timeout=self.timeout)
                    self.process_close()
                else:
                    self.process.wait()
            self.remove_instance(self)
            logger.info(f"{self.command} END")
        except Exception as e: 
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())
            logger.warning(self.command)
            self.send_stdout_callback(self.call_id, 'ERROR', str(e))
            self.send_stdout_callback(self.call_id, 'ERROR', str(traceback.format_exc()))
        finally:
            if self.stdout_callback != None:
                #self.stdout_callback(self.call_id, 'thread_end', None)
                pass
    

    def __start_communicate(self):
        self.stdout_queue = queue.Queue()
        sout = io.open(self.process.stdout.fileno(), 'rb', closefd=False)
       
        def Pump(stream):
            _queue = queue.Queue()
            
            def rdr():
                while True:
                    try:
                        buf = self.process.stdout.read(1)
                    except:
                        continue
                    #print(buf)
                    if buf:
                        _queue.put( buf )
                    else: 
                        _queue.put( None )
                        break
                _queue.put( None )
                time.sleep(1)

            def clct():
                active = True
                while active:
                    r = _queue.get()
                    if r is None:
                        break
                    try:
                        while True:
                            r1 = _queue.get(timeout=0.005)
                            if r1 is None:
                                active = False
                                break
                            else:
                                r += r1
                    except:
                        pass
                    if r is not None:
                        #print(f"{r=}")
                        self.stdout_queue.put(r)
                self.stdout_queue.put('\n')
                self.stdout_queue.put('<END>')
                self.stdout_queue.put('\n')
            for tgt in [rdr, clct]:
                th = threading.Thread(target=tgt)
                th.setDaemon(True)
                th.start()
        Pump(sout)


    def __start_send_callback(self):
        def func():
            while self.stdout_queue:
                line = self.stdout_queue.get()
                #logger.error(line)
                if line == '<END>':
                    self.send_stdout_callback(self.call_id, 'END', None)
                    break
                else:
                    self.send_stdout_callback(self.call_id, 'LOG', line)
            self.remove_instance(self)

        def func_callback_line():
            previous = ''
            while self.stdout_queue:
                receive = previous + self.stdout_queue.get()
                lines = receive.split('\n')
                previous = lines[-1]

                for line in lines[:-1]:
                    line = line.strip()
                    # TODO
                    #logger.error(line)
                    if line == '<END>':
                        self.send_stdout_callback(self.call_id, 'END', None)
                        break
                    else:
                        self.send_stdout_callback(self.call_id, 'LOG', line)
            self.remove_instance(self)

        if self.callback_line:
            th = threading.Thread(target=func_callback_line, args=())
        else:
            th = threading.Thread(target=func, args=())
        th.setDaemon(True)
        th.start()



    def process_close(self):
        try:
            if self.process is not None and self.process.poll() is None:
                #import psutil
                #process = psutil.Process(instance.process.pid)
                #for proc in instance.process.children(recursive=True):
                #    proc.kill()
                self.process.kill()
        except Exception as e: 
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())           
        finally:
            try:
                #self.stdout_queue = None 
                self.process.kill()
            except: pass
        
        self.remove_instance(self)

    def input_command(self, cmd):
        if self.process != None:
            self.process.stdin.write(f'{cmd}\n')
            self.process.stdin.flush()

    def send_stdout_callback(self, call_id, mode, data):
        try:
            if self.stdout_callback != None:
                self.stdout_callback(self.call_id, mode, data)
        except Exception as e: 
            logger.error(f'Exception:{str(e)}')
            logger.error(f"[{call_id}] [{mode}] [{data}]")
            #logger.error(traceback.format_exc())   


    @classmethod
    def all_process_close(cls):
        for instance in cls.__instance_list:
            instance.process_close()
        cls.__instance_list = []


    @classmethod
    def remove_instance(cls, remove_instance):
        new = []
        for instance in cls.__instance_list:
            if remove_instance.timestamp == instance.timestamp:
                continue
            new.append(instance)
        cls.__instance_list = new
    
    @classmethod
    def print(cls):
        for instance in cls.__instance_list:
            logger.info(instance.command)


    @classmethod
    def get_instance_by_call_id(cls, call_id):
        for instance in cls.__instance_list:
            if instance.call_id == call_id:
                return instance
    
    @classmethod
    def get_list(cls):
        return cls.__instance_list




import argparse
import os
import sys

sys.path.append(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'lib'))

from support import SupportFile, SupportSC, logger


class CodeEncode:
    def start_folder(self, folderpath):
        for name in os.listdir(folderpath):
            filepath = os.path.join(folderpath, name)
            if os.path.isfile(filepath) and name not in ['setup.py', '__init__.py'] and name.endswith('.py'):
                self.encode_file(filepath)

    def encode_file(self, filepath):
        if filepath.endswith(".py") == False:
            logger.info("is not .py file")
            return
        text = SupportFile.read_file(filepath)
        data = SupportSC.encode(text, 1)
        SupportFile.write_file(filepath + 'f', data)
        logger.info(f"Create {os.path.basename(filepath + 'f')}")


    def process_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--mode', default='encode')
        parser.add_argument('--source', required=True, help=u'absolute path. folder or file')
        args = parser.parse_args()
        if SupportSC.LIBRARY_LOADING == False:
            logger.error("sc import fail")
            return
        if os.path.exists(args.source):
            if os.path.isdir(args.source):
                self.start_folder(args.source)
            elif os.path.isfile(args.source):
                self.encode_file(args.source)
        else:
            logger.error("wrong source path!!")


if __name__== "__main__":
    CodeEncode().process_args()

# python C:\work\FlaskFarm\flaskfarm\cli\code_encode.py --source=C:\work\FlaskFarm\data\LOADING\sjva



import logging
import logging.handlers
import os
import platform
import shutil
import sys
import time
import traceback
from datetime import datetime

import yaml
from flask import Flask
from flask_cors import CORS
from flask_login import LoginManager, login_required
from flask_socketio import SocketIO
from flask_sqlalchemy import SQLAlchemy
from pytz import timezone, utc

from .init_declare import CustomFormatter, check_api


class Framework:
    __instance = None

    @classmethod
    def get_instance(cls):
        if cls.__instance == None:
            cls.__instance = Framework()
        return cls.__instance


    def __init__(self):
        self.logger = None
        self.app = None
        self.celery = None
        self.db = None
        self.scheduler = None
        self.socketio = None
        self.path_app_root = None
        self.path_data = None
        self.users = {}

        self.__level_unset_logger_list = []
        self.__logger_list = []
        self.all_log_filehandler = None
        self.__exit_code = -1
        self.login_manager = None
        #self.plugin_instance_list = {}
        #self.plugin_menus = {}
        
        # 그냥 F. 로 접근 하게....
        self.SystemModelSetting = None
        self.Job = None
        self.login_required = login_required
        self.check_api = check_api
        self.__initialize()
     

    def __initialize(self):
        os.environ["PYTHONUNBUFFERED"] = "1"
        os.environ['FF'] = "true"
        os.environ['FF_PYTHON'] = sys.executable
        self.__config_initialize("first")
        self.__make_default_dir()
        
        self.logger = self.get_logger(__package__)
        import support

        self.__prepare_starting()
        self.app = Flask(__name__)
        self.__config_initialize('flask')

        self.__init_db()
        
        if True or self.config['run_flask']:
            from .scheduler import Job, Scheduler
            self.scheduler = Scheduler(self)
            self.Job = Job

        if self.config['use_gevent']:
            self.socketio = SocketIO(self.app, cors_allowed_origins="*")
        else:
            self.socketio = SocketIO(self.app, cors_allowed_origins="*", async_mode='threading')
        
        CORS(self.app)

        self.login_manager = LoginManager()
        self.login_manager.init_app(self.app)
        self.login_manager.login_view = "/system/login"
        
        self.celery = self.__init_celery()
        
        from flask_dropzone import Dropzone
        self.app.config.update(
            DROPZONE_MAX_FILE_SIZE = 102400,
            DROPZONE_TIMEOUT = 5*60*1000,
            DROPZONE_ALLOWED_FILE_CUSTOM = True,
            DROPZONE_ALLOWED_FILE_TYPE = "image/*, audio/*, video/*, text/*, application/*, *.*",
        )
        self.dropzone = Dropzone(self.app)


    def __init_db(self):
        # https://flask-sqlalchemy.palletsprojects.com/en/3.0.x/config/#flask_sqlalchemy.config.SQLALCHEMY_BINDS
        # 어떤 편법도 불가. db를 사용하지 않아도 파일이 생김.
        db_path = os.path.join(self.config['path_data'], 'db', 'system.db')
        self.app.config['SQLALCHEMY_DATABASE_URI'] = f'sqlite:///{db_path}?check_same_thread=False' # 3.0에서 필수
        self.app.config['SQLALCHEMY_BINDS'] = {'system':f'sqlite:///{db_path}?check_same_thread=False'}
        self.app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

        _ = os.path.join(self.config['path_data'], 'plugins')
        plugins = []
        if os.path.exists(_):
            plugins = os.listdir(_)
        
        if self.config['path_dev'] != None:
            if type(self.config['path_dev']) == type(''):
                plugin_path_list = [self.config['path_dev']]
            elif type(self.config['path_dev']) == type([]):
                plugin_path_list = self.config['path_dev']
            for __ in plugin_path_list:
                if os.path.exists(__):
                    plugins += os.listdir(__)
        
        for package_name in plugins:
            db_path = os.path.join(self.config['path_data'], 'db', f'{package_name}.db')
            self.app.config['SQLALCHEMY_BINDS'][package_name] = f'sqlite:///{db_path}?check_same_thread=False'
        self.db = SQLAlchemy(self.app, session_options={"autoflush": False, "expire_on_commit": False})  
        #with self.app.app_context():
        #    self.db.session.expunge_all()

        
    def __init_celery(self):
        try:
            from celery import Celery

            #if frame.config['use_celery'] == False or platform.system() == 'Windows':
            if self.config['use_celery'] == False:
                raise Exception('no celery')
            
            redis_port = os.environ.get('REDIS_PORT', None)
            if redis_port == None:
                redis_port = self.config.get('redis_port', None)
            if redis_port == None:
                redis_port = '6379'
  

            self.app.config['CELERY_BROKER_URL'] = 'redis://localhost:%s/0' % redis_port
            self.app.config['CELERY_RESULT_BACKEND'] = 'redis://localhost:%s/0' % redis_port
            celery = Celery(self.app.name, broker=self.app.config['CELERY_BROKER_URL'], backend=self.app.config['CELERY_RESULT_BACKEND'])
            celery.conf['CELERY_ENABLE_UTC'] = False
            celery.conf.update(
                task_serializer='pickle',
                result_serializer='pickle',
                accept_content=['pickle'],
                timezone='Asia/Seoul'
            )
            from celery import bootsteps
            from click import Option
            celery.user_options['worker'].add(Option(('--config_filepath',), help=''),)
            celery.user_options['worker'].add(Option(('--running_type',), help=''),)
            class CustomArgs(bootsteps.Step):
                def __init__(self, worker, config_filepath=None, running_type=None,  **options):
                    from . import F
                    F.logger.info(f"celery config_filepath: {config_filepath}")
                    F.logger.info(f"celery running_type: {running_type}")
                    #F.logger.info(f"celery running_type: {options}")
            celery.steps['worker'].add(CustomArgs)
        except Exception as e:
            if self.config['use_celery']:
                self.logger.error('CELERY!!!')
                self.logger.error(f'Exception:{str(e)}')
                self.logger.error(traceback.format_exc())
            else:
                self.logger.info("use_celery = False")

            def dummy_func():
                pass
                    
            class celery(object):
                class task(object):
                    def __init__(self, *args, **kwargs):
                        if len(args) > 0:
                            self.f = args[0]
                
                    def __call__(self, *args, **kwargs):
                        if len(args) > 0 and type(args[0]) == type(dummy_func):
                            return args[0]
                        self.f(*args, **kwargs)
        return celery


    def initialize_system(self):
        from system.setup import P
        SystemInstance = P
        try:
            with self.app.app_context():
                self.db.create_all()
        except Exception as e:
            self.logger.error('CRITICAL db.create_all()!!!')
            self.logger.error(f'Exception:{str(e)}')
            self.logger.error(traceback.format_exc())
        self.SystemModelSetting = SystemInstance.ModelSetting 
        if self.config['run_flask']:
            SystemInstance.plugin_load()
        self.app.register_blueprint(SystemInstance.blueprint)
        self.config['flag_system_loading'] = True
        self.__config_initialize('member')
        self.__config_initialize('system_loading_after')
        self.set_level(self.SystemModelSetting.get_int('log_level'))
       

    def initialize_plugin(self): 
        from system.setup import P as SP

        from .init_web import jinja_initialize
        jinja_initialize(self.app)
        
        from .init_plugin import PluginManager
        self.PluginManager = PluginManager
        PluginManager.plugin_update()
        PluginManager.plugin_init()
        PluginManager.plugin_menus['system'] = {'menu':SP.menu, 'match':False} 

        from .init_menu import MenuManager
        MenuManager.init_menu()
        self.logger.info("메뉴 구성 완료")

        if self.config['run_flask']:
            if self.config.get('port') == None:
                self.config['port'] = self.SystemModelSetting.get_int('port')

        from . import init_route, log_viewer

        self.__make_default_logger()
        self.__config_initialize("last")
        self.logger.info('### LAST')
        self.logger.info(f"### PORT: {self.config.get('port')}")
        self.logger.info('### Now you can access App by webbrowser!!')


    def __prepare_starting(self):
        # 여기서 monkey.patch시 너무 늦다고 문제 발생
        pass
        
        
    ###################################################
    # 환경
    ###################################################
    def __config_initialize(self, mode):
        if mode == "first":
            self.config = {}
            self.config['os'] = platform.system()
            self.config['flag_system_loading'] = False
            #self.config['run_flask'] = True if sys.argv[0].endswith('main.py') else False
            self.config['run_celery'] = True if sys.argv[0].find('celery') != -1 else False
            self.config['run_flask'] = not self.config['run_celery']
            self.config['path_app'] = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
            if self.config['os'] == 'Windows' and self.config['path_app'][0] != '/':
                self.config['path_app'] = self.config['path_app'][0].upper() + self.config['path_app'][1:]
            self.path_app_root = self.config['path_app']
            self.config['path_working'] = os.getcwd()
            if os.environ.get('RUNNING_TYPE', None) != None:
                self.config['running_type'] = os.environ.get('RUNNING_TYPE')
            self.config['export_filepath'] = os.path.join(self.config['path_app'], 'export.sh')
            self.config['exist_export'] = os.path.exists(self.config['export_filepath'])
            self.config['recent_version'] = '--'
            from .version import VERSION
            self.config['version'] = VERSION
            self.__process_args()
            self.__load_config()
            self.__init_define()
            self.config['menu_yaml_filepath'] = os.path.join(self.config['path_data'], 'db', 'menu.yaml')
            self.config['notify_yaml_filepath'] = os.path.join(self.config['path_data'], 'db', 'notify.yaml')
            if 'running_type' not in self.config:
                self.config['running_type'] = 'native'
            self.pip_install()
        elif mode == "flask":
            self.app.secret_key = os.urandom(24)
            self.app.config['TEMPLATES_AUTO_RELOAD'] = True
            self.app.config['JSON_AS_ASCII'] = False
        elif mode == 'system_loading_after':
            pass
        elif mode == 'last':
            db_foder = os.path.join(self.config['path_data'], 'db')
            for name in os.listdir(db_foder):
                if name.endswith('.db'):
                    db_filepath = os.path.join(db_foder, name)
                    try:
                        if os.stat(db_filepath).st_size == 0:
                            os.remove(db_filepath)
                            self.logger.debug(f"REMOVE {db_filepath}")
                    except:
                        pass



    def __init_define(self):
        self.config['DEFINE'] = {}
        # 이건 필요 없음
        self.config['DEFINE']['GIT_VERSION_URL'] = 'https://raw.githubusercontent.com/flaskfarm/flaskfarm/main/lib/framework/version.py'
        self.config['DEFINE']['CHANGELOG'] = 'https://github.com/flaskfarm/flaskfarm'
        self.config['DEFINE']['WEB_DIRECT_URL'] = "http://52.78.103.230:49734"


    def __process_args(self):
        # celery 에서 args 처리시 문제 발생.
        if self.config['run_flask']:
            import argparse
            parser = argparse.ArgumentParser()
            parser.add_argument('--config', default='.', help='config filepath. Default: {current folder}/config.yaml')
            parser.add_argument('--repeat', default=0, type=int, help=u'Do not set. This value is set by automatic')
            args = parser.parse_args()
            self.config['arg_repeat'] = args.repeat
            self.config['arg_config'] = args.config
        else:
            # 아주 안좋은 구조..
            # celery user_options으로 configfilepath를 받은 후 처리해야하나, 로그파일 경로 등에서 데이터 폴더 위치를 미리 사용하는 경우가 많다.
            # sys.argv에서 데이터 경로를 바로 가져와서 사용.
            self.config['arg_repeat'] = 0
            #self.config['arg_config'] = sys.argv[-1].split('=')[-1]
            #self.config['arg_config'] = sys.argv[-1].split('=')[-1]
            for tmp in sys.argv:
                if tmp.startswith('--config_filepath'):
                    self.config['arg_config'] = tmp.split('=')[1]
                    #break
                elif tmp.startswith('--running_type'):
                    self.config['running_type'] = tmp.split('=')[1]

            #self.config['arg_config'] = 

    def __load_config(self):
        #if self.config['run_flask']:
        if self.config['arg_config'] == '.':
            #self.config['config_filepath'] = os.path.join(self.path_app_root, 'config.yaml')
            self.config['config_filepath'] = os.path.join(self.config['path_working'], 'config.yaml')
        else:
            self.config['config_filepath'] = self.config['arg_config']
        if os.path.exists(self.config['config_filepath']) == False:
            if self.config.get('running_type', '').startswith('docker'):
                with open(self.config['config_filepath'], 'w', encoding='utf8') as f:
                    yaml.dump({'path_data':'/data'}, f, default_flow_style=False, allow_unicode=True)
            else:                
                shutil.copy(
                    os.path.join(self.path_app_root, 'files', 'config.yaml.template'),
                    self.config['config_filepath']
                )
        with open(self.config['config_filepath'], encoding='utf8') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        for key, value in data.items():
            if key == 'running_type' and value not in ['termux', 'entware']:
                continue
            self.config[key] = value

        if self.config['path_data'] == '.':
            self.config['path_data'] = self.config['path_working']
            # 예외적으로 현재폴더가 app일 경우 지저분해지는 것을 방지하기 위해 data 로 지정
            if self.config['path_data'] == self.config['path_working']:
                self.config['path_data'] = os.path.join(self.config['path_working'], 'data')
        self.path_data = self.config['path_data'] 

        if self.config.get('use_gevent') == None:
            self.config['use_gevent'] = True
        if self.config.get('use_celery') == None:
            self.config['use_celery'] = True
        if self.config.get('debug') == None:
            self.config['debug'] = False
        if self.config.get('plugin_update') == None:
            self.config['plugin_update'] = True
        # 2022-11-20
        if self.config['debug']:
            self.config['plugin_update'] = False
        if self.config.get('plugin_loading_only_devpath') == None:
            self.config['plugin_loading_only_devpath'] = False
        if self.config.get('plugin_loading_list') == None:
            self.config['plugin_loading_list'] = []
        if self.config.get('plugin_except_list') == None:
            self.config['plugin_except_list'] = []
        if self.config.get('path_dev') == None:
            self.config['path_dev'] = None
        

    def __make_default_dir(self):
        os.makedirs(self.config['path_data'], exist_ok=True)
        tmp = os.path.join(self.config['path_data'], 'tmp')
        try:
            import shutil
            if os.path.exists(tmp):
                shutil.rmtree(tmp)
        except:
            pass
        sub = ['db', 'log', 'tmp']
        for item in sub:
            tmp = os.path.join(self.config['path_data'], item)
            os.makedirs(tmp, exist_ok=True)
        
    ###################################################



    
    ###################################################
    # 로그
    ###################################################
    def get_logger(self, name, from_command=False):
        logger = logging.getLogger(name)
        if not logger.handlers:
            level = logging.DEBUG
            try:
                if self.config['flag_system_loading']:
                    try:
                        #from system import SystemModelSetting
                        level = self.SystemModelSetting.get_int('log_level')
                    except:
                        level = logging.DEBUG
                    if self.__level_unset_logger_list is not None:
                        for item in self.__level_unset_logger_list:
                            item.setLevel(level)
                        self.__level_unset_logger_list = None
                else:
                    self.__level_unset_logger_list.append(logger)
                if name.startswith('apscheduler'):
                    level = logging.CRITICAL
                else:
                    self.__logger_list.append(logger)
            except:
                pass
            logger.setLevel(level)
            def customTime(*args):
                utc_dt = utc.localize(datetime.utcnow())
                my_tz = timezone("Asia/Seoul")
                converted = utc_dt.astimezone(my_tz) 
                return converted.timetuple()

            if from_command == False:
                file_formatter = logging.Formatter(u'[%(asctime)s|%(levelname)s|%(name)s|%(filename)s:%(lineno)s] %(message)s')
            else:
                file_formatter = logging.Formatter(u'[%(asctime)s] %(message)s')

            file_formatter.converter = customTime
            file_max_bytes = 1 * 1024 * 1024 
            fileHandler = logging.handlers.RotatingFileHandler(filename=os.path.join(self.path_data, 'log', f'{name}.log'), maxBytes=file_max_bytes, backupCount=5, encoding='utf8', delay=True)
            fileHandler.setFormatter(file_formatter) 
            logger.addHandler(fileHandler)
            if name == 'framework' and self.all_log_filehandler == None:
                self.all_log_filehandler = logging.handlers.RotatingFileHandler(filename=os.path.join(self.path_data, 'log', f'all.log'), maxBytes=5*1024*1024, backupCount=5, encoding='utf8', delay=True)
                self.all_log_filehandler.setFormatter(file_formatter) 
            
            if from_command == False:
                streamHandler = logging.StreamHandler() 
                streamHandler.setFormatter(CustomFormatter()) 
                logger.addHandler(streamHandler)
            
            if self.all_log_filehandler != None:
                logger.addHandler(self.all_log_filehandler)
            
        return logger 


    def __make_default_logger(self):
        self.get_logger('apscheduler.scheduler')
        self.get_logger('apscheduler.executors.default')
        try: logging.getLogger('socketio').setLevel(logging.ERROR)
        except: pass
        try: logging.getLogger('engineio').setLevel(logging.ERROR)
        except: pass
        try: logging.getLogger('apscheduler.scheduler').setLevel(logging.ERROR)
        except: pass
        try: logging.getLogger('apscheduler.executors.default').setLevel(logging.ERROR)
        except: pass
        try: logging.getLogger('werkzeug').setLevel(logging.ERROR)
        except: pass

    def set_level(self, level):
        try:
            for l in self.__logger_list:
                l.setLevel(int(level))
            self.__make_default_logger()
        except:
            pass    
    ###################################################
    

    def start(self):
        host = '0.0.0.0'
        for i in range(5): 
            try: 
                #self.logger.debug(d(self.config))
                # allow_unsafe_werkzeug=True termux  nohup 실행시 필요함
                #if self.config['running_type'] == 'termux':
                #    self.socketio.run(self.app, host=host, port=self.config['port'], debug=self.config['debug'], use_reloader=self.config['debug'], allow_unsafe_werkzeug=True)
                #else:
                self.socketio.run(self.app, host=host, port=self.config['port'], debug=self.config['debug'], use_reloader=self.config['debug'])
                self.logger.warning(f"EXIT CODE : {self.__exit_code}")
                # 2021-05-18  
                if self.config['running_type'] in ['termux', 'entware']:
                    os._exit(self.__exit_code)
                else:
                    if self.__exit_code != -1:
                        sys.exit(self.__exit_code)
                    else:
                        self.logger.warning(f"framework.exit_code is -1")
                break
            except Exception as exception:
                self.logger.error(f"Start ERROR : {str(exception)}")
                host = '127.0.0.1'  
                time.sleep(10*i)
                continue  
            except KeyboardInterrupt: 
                self.logger.error('KeyboardInterrupt !!')
            
                
            #except SystemExit: 
            #    return 
                #sys.exit(self.__exit_code)


    # system 플러그인에서 콜
    def restart(self):
        self.__exit_code = 1
        self.__app_close()

    def shutdown(self):
        self.__exit_code = 0
        self.__app_close()

    def __app_close(self):
        try:
            from support import SupportSubprocess
            SupportSubprocess.all_process_close()
            from .init_plugin import PluginManager
            PluginManager.plugin_unload()
            with self.app.test_request_context():
                self.socketio.stop()
        except Exception as e: 
            self.logger.error(f"Exception:{str(e)}")
            self.logger.error(traceback.format_exc())

    def get_recent_version(self):
        try:
            import requests
            text = requests.get(self.config['DEFINE']['GIT_VERSION_URL']).text
            self.config['recent_version'] = text.split('=')[1].strip().strip('"')
            return True
        except Exception as e:
            self.logger.error(f'Exception:{str(e)}')
            self.logger.error(traceback.format_exc())
            self.config['recent_version'] =  '확인 실패'
        return False        


    # dev 도커용. package는 setup에 포함.
    def pip_install(self):
        try:
            import json_fix
        except:
            os.system('pip install json_fix')



__menu = {
    'uri' : __package__,
    'name': '설정',
    'list': [
        {
            'uri': 'setting',
            'name': '일반설정',
            'list': [
                {'uri': 'basic', 'name': '기본'},
                {'uri': 'auth', 'name': '인증'},
                {'uri': 'web', 'name': '웹'},
                {'uri': 'menu', 'name': '메뉴 구성'},                
                {'uri': 'config', 'name': 'config.yaml 파일'},
                {'uri': 'export', 'name': 'export.sh 파일'},
                {'uri': 'celery', 'name': '비동기 작업(celery)'},
                {'uri': 'notify', 'name': '알림'},
                {'uri': 'manual/files/매뉴얼_설정.md', 'name': '매뉴얼'},
            ],
        },
        {
            'uri': 'plugin', 
            'name': '플러그인',
            'list': [
                #{'uri': 'setting', 'name': '설정'},
                {'uri': 'list', 'name': '로딩 플러그인'},
                {'uri': 'all', 'name': '전체 플러그인 목록'},
            ],
        },
        {
            'uri': 'tool',
            'name': '시스템 툴',
            'list': [
                {'uri': 'command', 'name': 'Command'},
                {'uri': 'upload', 'name': '업로드'},
                {'uri': 'python', 'name': 'Python'},
                #{'uri': 'db', 'name': 'DB'},
                #{'uri': 'crypt', 'name': '암호화'},
            ]
        },
        {
            'uri': 'all_log',
            'name': '전체 로그'
        }
    ]
}


import os

from framework import F

export = os.path.join(F.config['path_app'], 'export.sh')
if os.path.exists(export) == False:
    for mod in __menu['list']:
        if mod['uri'] == 'setting':
            del mod['list'][5]




setting = {
    'filepath' : __file__,
    'use_db': True,
    'use_default_setting': True,
    'home_module': 'setting',
    'menu': __menu,
    'setting_menu': None,
    'default_route': 'normal',
}

try:
    from plugin import *
    P = create_plugin_instance(setting)

    SystemModelSetting = P.ModelSetting
    from .mod_home import ModuleHome
    from .mod_log import ModuleLog
    from .mod_plugin import ModulePlugin
    from .mod_route import ModuleRoute
    from .mod_setting import ModuleSetting
    from .mod_tool import ModuleTool
    P.set_module_list([ModuleHome, ModuleRoute, ModuleSetting, ModulePlugin, ModuleTool, ModuleLog])

except Exception as e:
    F.logger.error(f'Exception:{str(e)}')
    F.logger.error(traceback.format_exc())



# 순서 바꾸지 말 것
import os, sys, traceback, re, threading, time, queue, json, shutil, yaml
import requests
from datetime import datetime, timedelta
from flask import Blueprint, render_template, jsonify, redirect, request
from sqlalchemy import desc, or_

from framework import logger
from .model_setting import get_model_setting
from .logic import Logic
from .route import default_route, default_route_socketio_module, default_route_socketio_page, default_route_single_module
from .logic_module_base import PluginModuleBase, PluginPageBase
#from .ffmpeg_queue import FfmpegQueueEntity, FfmpegQueue
from .model_base import ModelBase
from .create_plugin import create_plugin_instance

from framework import *
from tool import *



import os
import random
import time
import traceback

import requests

try:
    from discord_webhook import DiscordEmbed, DiscordWebhook
except:
    os.system('pip3 install discord-webhook')

# 2023-10-13 by flaskfarm
# 웹훅 URL이 git에 노출되면 중단.
# base64로 인코딩.
import base64

from discord_webhook import DiscordEmbed, DiscordWebhook

from . import logger

webhook_list = [
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyODM4MzE4MzI4MjIyNy9ORXpNWFBmT05vbUU3bl8xck1iT0ZWQUI4ZmlXN21vRFlGYnJHRk03UlJSWF90ZGMyS0lxY2hWcXV6VF8wVm5ZUEJRVQ==').decode('utf-8'), # 1
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyODc5MzExMzU4MzY3Ni94dEZlWnRWbkhEUGc4aFBYZkZDMkFidUtDSmlwNjQ0d1RQMDFJalVncTR5ODB6XzZCRi1kTFctemlEdGNlWF84RXVtRw==').decode('utf-8'), # 2
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyODkyMTA4MTgxMDk1NS9xOUVYZndHYll6bHdwM1MtMnpxUmxZcnJYWS1nTUttTTRlTUd0YW8zNTF1d1c2N0U2ckNFUW0zWDJhbDJURnFXMHR4cw==').decode('utf-8'), # 3
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTA4MzA0NDg1MTc3My9jVU1YRkVERHQ2emtWOW90Mmd5dlpYeVlOZV9VcGtmcmZhTzg5aHZoLVdod0c0Z24zOGJhT19DVkg2Z0N4clFraVZRcA==').decode('utf-8'), # 4
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTE4MzAwMzQ5MjM5Mi9CVXl1U3lKTHc1cktHdFRKOWhqRDk3SklKWW9HSTZ6SnJ4MzdLX0s4TkVKU3ZTYlZ4aC0tMVFRMEFZbTJFa0tzaEJRcQ==').decode('utf-8'), # 5
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTI4OTQ4ODQ5NDY0My9XLWhZck95QTBza2M1dUdVTkpyenU4ZHFSMVF0QmMtOVAzMW45RHhQWkhVLXptdEZ3MWVLWTE0dlZubkRUV25EU2ZRTw==').decode('utf-8'), # 6
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTQwOTExODQzNzQ2OS95NDRjVTkwM1hLS2NyaWFidERHMzRuMzZfRkZsMF9TV2p4b0lWMlBZY0dxNWxyU1dxVWt5ZklkZlcwM0FFVDJObThMaQ==').decode('utf-8'), # 7
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTQ5NTQyNDYxODQ5Ni9PSnFlVHRhZ1FtVGFrU2VNQkRucVhZRTJieWRuX2cweHV2VTA0WmdKS3NjWEQydkVHbHBWYzdhQWZMQ0ZYSXNTbTN5OQ==').decode('utf-8'), # 8
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTU4NzE2NjY0MjM2OC9PeG9NRllUT1dOWmcwS3pfQ3k2ZXowX2laQnpqbm02QVkyTjJfMWhaVHhYdWxPRm5oZ0lsOGxxcUVYNVVWRWhyMHJNcw==').decode('utf-8'), # 9
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTY4OTM3NzYzNjQ3Mi9iblgyYTNsWjI1R2NZZ1g4Sy1JeXZVcS1TMV9zYmhkbEtoSTI4eWp6SWdHOHRyekFWMXBKUkgxYkdrdmhzMVNUNS1uMg==').decode('utf-8'), # 10
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTgwMTg0MzY5MTY4Mi96OUFocjZjNmxaS1VyWV9TRmhfODVQeEVlSjJuQW8wMXlHS3RCUWpNNnJmR3JGVXdvQ1oyQ3NJYmlTMHQ1NDZwU3NUUg==').decode('utf-8'), # 11
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIyOTk0NjQyMTM2Njg0NC9BYnFVRlJCN0dzb3ktUkdfMzBLNXZkNm9XUWRnbkpDZ1ctTlpBbkFRN0N2TzdsNjRfeXRkY0ZHUkhLV2RldE1jQzhTSw==').decode('utf-8'), # 12
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDExNjAyNjQyMTM0OC82dG1NOTA2eV9QTHJ3WGFxcGNZS25OMEJIQjlDTkxJT1dJeTdpc3Exbm9VMHJxU2V0NzI2R1Y4Zk9Ua2pCbDZacXMxVA==').decode('utf-8'), # 13
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDIwMTExMjA3MjM2My9DYkdkcVdvd3hCcTV3ck1hck0taGZqajVIbFJ2VFFWa0tuZUVaVl9yMlc1UkxHZFZpWW15VzZZcl9PbEJCZG5KWk1wNw==').decode('utf-8'), # 14
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDI5MTUwNzcyNDQwOC9UanJFc08zSTJyT3l0d0ZvVFhUSlNTOXphaDJpbG9CcVk3TzhHMHZWbDROTmI0aGpaaDNjVGo0cGNla2lxa3RqaGRPTg==').decode('utf-8'), # 15
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDM3NDQ4NzgzNDc3NC85ZFpodjZfajRuT0hpbGtMaFVVc1B6OVFKa2dqQ3BJZ19PWE55YnZtV1BiQlNVdmRZWC1IVW5UM3RneDlKdnZlYjVMZw==').decode('utf-8'), # 16
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDU2ODkzMTU2OTc3NC82MTFBVTg0ZUZBcXllWlktQ2lPbnozbm4zSHg3ZldwQUNCbjlMTFNENUFJdHRkYjVHSm9pV3B1dHpxdEVHZ3l0RHlXYg==').decode('utf-8'), # 17
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDY3MzkxMDgwMDM4Ni9jbEZvZHEwREhGNUlvYUtVRXVRcXNGbnB3OXZoZUx1RU1qbVJFNjQyRUZGa21wYXBwMzhYWDNPMmZKWUVSdjMzY0tORg==').decode('utf-8'), # 18
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDc0OTMyMDE5NjE5OC8weE1vZ2o5UXRCM1NGZE5KYk04STk1LU9XQzI2Zm1WTWpjelpSX2REY2hnblZoUk1QelVzRHFlYTc0QUdISFRFVWFVZQ==').decode('utf-8'), # 19
    base64.b64decode(b'aHR0cHM6Ly9kaXNjb3JkLmNvbS9hcGkvd2ViaG9va3MvMTE2MjIzMDg0NTMxNTIyMzU2Mi9HWVItZUo2WFltdy1VMHRkWWY5QXdnU2JVZFpSb1k4aWx2MlVxaVJBSnlBdDBsYWV5ck1tSzJmRGp5T1YyenBJUUR2bA==').decode('utf-8'), # 20
]


class SupportDiscord(object):

    @classmethod
    def send_discord_message(cls, text, image_url=None, webhook_url=None):
        try:
            """
            webhook = DiscordWebhook(url=webhook_url, content=text)
            if image_url is not None:
                embed = DiscordEmbed()
                embed.set_timestamp()
                embed.set_image(url=image_url)
                webhook.add_embed(embed)
            """
            try:
                
                if image_url is not None:
                    webhook = DiscordWebhook(url=webhook_url)
                    embed = DiscordEmbed()
                    embed.set_timestamp()
                    embed.set_image(url=image_url)
                    tmp = text.split('\n', 1)
                    embed.set_title(tmp[0])
                    embed.set_description(tmp[1])
                    webhook.add_embed(embed)
                else:
                    if 'http://' in text or 'https://' in text:
                        webhook = DiscordWebhook(url=webhook_url, content= text)
                    else:
                        webhook = DiscordWebhook(url=webhook_url, content='```' + text + '```')
                webhook.execute()
                return True
            except:
                webhook = DiscordWebhook(url=webhook_url, content=text)
                if image_url is not None:
                    embed = DiscordEmbed()
                    embed.set_timestamp()
                    embed.set_image(url=image_url)
                    webhook.add_embed(embed)
                webhook.execute()
                return True
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 
        return False


    @classmethod
    def send_discord_bot_message(cls, text, webhook_url, encryped=True):
        try:
            from support import SupportAES
            if encryped:
                text = '^' + SupportAES.encrypt(text)
            return cls.send_discord_message(text, webhook_url=webhook_url)
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc()) 
            return False


    
    @classmethod
    def discord_proxy_image(cls, image_url, webhook_url=None, retry=True):
        #2020-12-23
        #image_url = None
        if image_url == '' or image_url is None:
            return
        data = None
        
        if webhook_url is None or webhook_url == '':
            webhook_url =  webhook_list[random.randint(0,len(webhook_list)-1)]

        try:
            webhook = DiscordWebhook(url=webhook_url, content='')
            embed = DiscordEmbed()
            embed.set_timestamp()
            embed.set_image(url=image_url)
            webhook.add_embed(embed)
            import io
            byteio = io.BytesIO()
            webhook.add_file(file=byteio.getvalue(), filename='dummy')
            response = webhook.execute()
            data = None
            if type(response) == type([]):
                if len(response) > 0:
                    data = response[0].json()
            else:
                data = response.json()    
            
            if data is not None and 'embeds' in data:
                target = data['embeds'][0]['image']['proxy_url']
                if requests.get(target).status_code == 200:
                    return target
                else:
                    return image_url
            else:
                raise Exception(str(data))
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image(image_url, webhook_url=None, retry=False)
            else:
                return image_url

    
    @classmethod
    def discord_proxy_image_localfile(cls, filepath, retry=True):
        data = None
        webhook_url =  webhook_list[random.randint(0,len(webhook_list)-1)]

        try:
            webhook = DiscordWebhook(url=webhook_url, content='')
            import io
            with open(filepath, 'rb') as fh:
                byteio = io.BytesIO(fh.read())
            webhook.add_file(file=byteio.getvalue(), filename='image.jpg')
            embed = DiscordEmbed()
            embed.set_image(url="attachment://image.jpg")
            response = webhook.execute()
            data = None
            if type(response) == type([]):
                if len(response) > 0:
                    data = response[0].json()
            else:
                data = response.json()    
            
            if data is not None and 'attachments' in data:
                target = data['attachments'][0]['url']
                if requests.get(target).status_code == 200:
                    return target
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_localfile(filepath, retry=False)
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())
    
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_localfile(filepath, retry=False)


    @classmethod
    def discord_proxy_image_bytes(cls, bytes, retry=True, format='jpg'):
        data = None
        idx = random.randint(0,len(webhook_list)-1)
        webhook_url =  webhook_list[idx]
        try:
            webhook = DiscordWebhook(url=webhook_url, content='')
            webhook.add_file(file=bytes, filename=f'image.{format}')
            embed = DiscordEmbed()
            embed.set_image(url=f"attachment://image.{format}")
            response = webhook.execute()
            data = None
            if type(response) == type([]):
                if len(response) > 0:
                    data = response[0].json()
            else:
                data = response.json()    
            if data is not None and 'attachments' in data:
                target = data['attachments'][0]['url']
                if requests.get(target).status_code == 200:
                    return target
            logger.error(f"discord webhook error : {webhook_url}")
            logger.error(f"discord webhook error : {idx}")
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_bytes(bytes, retry=False)
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())

            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_bytes(bytes, retry=False)




    # RSS에서 자막 올린거
    @classmethod    
    def discord_cdn(cls, byteio=None, filepath=None, filename=None, webhook_url="https://discord.com/api/webhooks/1050549730964410470/ttge1ggOfIxrCSeTmYbIIsUWyMGAQj-nN6QBgwZTqLcHtUKcqjZ8wFWSWAhHmZne57t7", content='', retry=True):
        data = None
        if webhook_url is None:
            webhook_url =  webhook_list[random.randint(0,9)] 

        try:
            webhook = DiscordWebhook(url=webhook_url, content=content)
            if byteio is None and filepath is not None:
                import io
                with open(filepath, 'rb') as fh:
                    byteio = io.BytesIO(fh.read())
                
            webhook.add_file(file=byteio.getvalue(), filename=filename)
            embed = DiscordEmbed()
            response = webhook.execute()
            data = None
            if type(response) == type([]):
                if len(response) > 0:
                    data = response[0].json()
            else:
                data = response.json()    
            
            if data is not None and 'attachments' in data:
                target = data['attachments'][0]['url']
                if requests.get(target).status_code == 200:
                    return target
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_localfile(filepath, retry=False)
        except Exception as e: 
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())
            if retry:
                time.sleep(1)
                return cls.discord_proxy_image_localfile(filepath, retry=False)
                



import queue
import shlex

from support import SupportSubprocess
from tool import ToolModalCommand

from .setup import *


class PageCommand(PluginPageBase):
    
    def __init__(self, P, parent):
        super(PageCommand, self).__init__(P, parent, name='command')
        self.db_default = {
            f'{self.parent.name}_{self.name}_recent': '',
        } 

    def process_menu(self, req):
        arg = self.P.ModelSetting.to_dict()
        arg['path_data'] = F.config['path_data']
        return render_template(f'{self.P.package_name}_{self.parent.name}_{self.name}.html', arg=arg)
        


    def process_command(self, command, arg1, arg2, arg3, req):
        ret = {'ret':'success'}
        if command == 'foreground_command':
            P.ModelSetting.set(f'{self.parent.name}_{self.name}_recent', arg1)
            self.__foreground_execute(arg1, shlex.split(arg1))
            
            return jsonify('')
        elif command == 'job_new':
            db_item = ModelCommand.job_new(arg1)
            ret['msg'] = f"ID:{db_item.id} 작업을 생성하였습니다." 
        elif command == 'job_list':
            ret['data'] = ModelCommand.job_list()
            
        elif command == 'job_save':
            data = P.logic.arg_to_dict(arg1)
            db_item = ModelCommand.get_by_id(data['job_id'])
            db_item.set_command(data['job_command'])
            db_item.args = data['job_command_args']
            db_item.description = data['job_description']
            db_item.schedule_mode = data['job_schedule_mode']
            db_item.schedule_auto_start = (data.get('job_schedule_auto_start', 'False') == 'True')
            db_item.schedule_interval = data.get('job_schedule_interval', '')
            db_item.save()
            ret['msg'] = '수정하였습니다.'
        elif command == 'job_remove':
            if ModelCommand.delete_by_id(arg1):
                ret['msg'] = '삭제하였습니다.'
            else:
                ret['ret'] = 'danger'
                ret['msg'] = '삭제에 실패하였습니다.'
        elif command == 'job_fore_execute':
            db_item = ModelCommand.get_by_id(arg1)
            cmd = (db_item.command + ' ' + db_item.args).strip()
            self.__foreground_execute(f"Command ID: {db_item.id}", shlex.split(cmd), db_item.id)
        elif command == 'job_back_execute':
            self.execute_thread_start(arg1)
            ret['msg'] = "실행 요청을 하였습니다.<br>로그를 확인하세요."
        elif command == 'job_log':
            ret['filename'] = f"command_{arg1}.log"
            if os.path.exists(os.path.join(F.config['path_data'], 'log', f"command_{arg1}.log")) == False:
                ret['ret'] = 'danger'
                ret['msg'] = "로그 파일이 없습니다."
        elif command == 'task_sched':
            job_id = arg1
            flag = (arg2 == 'true')
            scheduler_id = f'command_{job_id}'
            if flag and F.scheduler.is_include(scheduler_id):
                ret['msg'] = '이미 스케쥴러에 등록되어 있습니다.'
            elif flag and F.scheduler.is_include(scheduler_id) == False:
                result = self.__sched_add(job_id)
                ret['msg'] = '스케쥴러에 추가하였습니다.'
            elif flag == False and scheduler.is_include(scheduler_id):
                result = scheduler.remove_job(scheduler_id)
                ret['msg'] = '스케쥴링 취소'
            elif flag == False and scheduler.is_include(scheduler_id) == False:
                ret['msg'] = '등록되어 있지 않습니다.'
        elif command == 'job_process_stop':
            process_ins = SupportSubprocess.get_instance_by_call_id(f"command_{arg1}")
            if process_ins == None:
                ret['msg'] = "실행중인 Process가 없습니다."
            else:
                process_ins.process_close()
                ret['msg'] = "Process를 중지하였습니다."
        return jsonify(ret)

    
    def __foreground_execute(self, title, command, job_id=None):
        
        if command[0] != 'LOAD':
            ToolModalCommand.start(title, [command])
        else:
            F.socketio.emit("command_modal_show", title, namespace='/framework')
            def start_communicate_load(load_log_list):
                def func():
                    while True:
                        logs = load_log_list.getvalue()
                        load_log_list.truncate(0)
                        if logs:
                            P.logger.error(logs)
                            F.socketio.emit("command_modal_add_text", logs.strip() + '\n', namespace='/framework')
                            if logs == '<<END>>':
                                break
                        time.sleep(0.3)
                th = threading.Thread(target=func)
                th.setDaemon(True)
                th.start()

            def func():
                import io
                from contextlib import redirect_stdout
                load_log_list = io.StringIO()
                with redirect_stdout(load_log_list):
                    start_communicate_load(load_log_list)
                    if job_id is not None:
                        command_logger = get_logger(f'command_{job_id}')
                    else:
                        command_logger = P.logger
                    self.__module_load(command, logger=command_logger)
                load_log_list.write("<<END>>")
                load_log_list.flush()
            th = threading.Thread(target=func, args=())
            th.setDaemon(True)
            th.start()
            return 'success'


    def __module_load(self, command, **kwargs):
        try:
            python_filename = command[1]
            python_sys_path = os.path.dirname(python_filename)
            if python_sys_path not in sys.path:
                sys.path.append(python_sys_path)
            module_name = os.path.basename(python_filename).split('.py')[0]
            if module_name not in sys.path:
                sys.path.append(module_name)
            import importlib
            mod = importlib.import_module(module_name)
            importlib.reload(mod)
            args = command
            mod_command_load = getattr(mod, 'main')
            if mod_command_load:
                ret = mod_command_load(*args, **kwargs)
            return ret
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())   



    def execute_thread_start(self, job_id):
        th = threading.Thread(target=self.execute_thread_function_by_job_id, args=(job_id,))
        th.setDaemon(True)
        th.start()
        return th


    def execute_thread_function_by_job_id(self, *args, **kwargs):
        #P.logger.error(d(args))
        #P.logger.error(d(kwargs))
        db_item = ModelCommand.get_by_id(args[0])
        kwargs['id'] = args[0]
        self.execute_thread_function((db_item.command + ' ' + db_item.args).strip(), **kwargs)

    
    def execute_thread_function(self, command, **kwargs):
        try:
            cmd = shlex.split(command)
            
            if cmd[0] == 'LOAD':
                command_logger = F.get_logger(f"command_{kwargs['id']}")
                kwargs['logger'] = command_logger
                return self.__module_load(cmd, **kwargs)
            else:
                class LogReceiver:
                    def __init__(self, logger):
                        self.logger = logger

                    def stdout_callback(self, call_id, mode, text):
                        if mode == 'LOG':
                            self.logger.debug(text)
                        else:
                            self.logger.debug(mode)
                command_logger = F.get_logger(f"command_{kwargs['id']}", from_command=True)
                receiver = LogReceiver(command_logger)
                process = SupportSubprocess(cmd, stdout_callback=receiver.stdout_callback, call_id=f"command_{kwargs['id']}")
                process.start()
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())        


    def plugin_load(self):
        def plugin_load_thread():
            try:
                db_items = ModelCommand.get_list()
                for db_item in db_items:
                    if db_item.schedule_mode == 'startup':
                        self.execute_thread_start(db_item.id)
                    elif db_item.schedule_mode == 'scheduler' and db_item.schedule_auto_start:
                        self.__sched_add(db_item.id, db_item=db_item)
            except Exception as e: 
                logger.error(f"Exception:{str(e)}")
                logger.error(traceback.format_exc())        
        try:
            th = threading.Thread(target=plugin_load_thread)
            th.setDaemon(True)
            th.start()
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())   
    
    
    def __sched_add(self, id, db_item=None):
        try:
            if db_item is None:
                db_item = ModelCommand.get_by_id(id)
            job_id = f"command_{db_item.id}"
            if scheduler.is_include(job_id):
                return
            job = Job(self.P.package_name, job_id, db_item.schedule_interval, self.execute_thread_function_by_job_id, db_item.description,  args=(db_item.id,))
            scheduler.add_job_instance(job)
            return True
        except Exception as e: 
            P.logger.error(f'Exception:{str(e)}')
            P.logger.error(traceback.format_exc())   
        return False










class ModelCommand(ModelBase):
    __tablename__ = 'command_job'
    __table_args__ = {'mysql_collate': 'utf8_general_ci'}
    __bind_key__ = 'system'

    id = db.Column(db.Integer, primary_key=True)
    command = db.Column(db.String)
    filepath = db.Column(db.String)
    args = db.Column(db.String)
    description = db.Column(db.String)
    schedule_mode = db.Column(db.String) # none, startup, scheduler
    schedule_auto_start = db.Column(db.Boolean)  # 시작시 스케쥴링 등록
    schedule_interval = db.Column(db.String)  # 주기
    

    def __init__(self, command):
        self.args = ''
        self.description = ''
        self.schedule_mode = 'none'
        self.schedule_auto_start = False
        self.schedule_interval = ''
        self.set_command(command)

    def set_command(self, command):
        self.command = command
        tmp = command.split(' ')
        for t in tmp:
            for ext in ['.py', '.sh', '.bat']:
                if t.endswith(ext):
                    self.filepath = t
                    break


    @classmethod
    def job_new(cls, command):
        item = ModelCommand(command)
        return item.save()


    @classmethod
    def job_list(cls):
        try:
            data = cls.get_list(by_dict=True)
            for item in data:
                item['scheduler_is_include'] = F.scheduler.is_include(f"command_{item['id']}")
                item['scheduler_is_running'] = F.scheduler.is_running(f"command_{item['id']}")

                item['process'] = (SupportSubprocess.get_instance_by_call_id(f"command_{item['id']}") != None)

            return data
        except Exception as e:
            logger.error(f"Exception:{str(e)}")
            logger.error(traceback.format_exc())



import abc
import os
import queue
import threading
import time
import traceback
from datetime import datetime


class FfmpegQueueEntity(abc.ABCMeta('ABC', (object,), {'__slots__': ()})):

    def __init__(self, P, module_logic, info):
        self.P = P
        self.module_logic = module_logic
        self.entity_id = -1 #FfmpegQueueEntity.static_index
        self.info = info
        self.url = None
        self.ffmpeg_status = -1
        self.ffmpeg_status_kor = u'대기중'
        self.ffmpeg_percent = 0
        self.ffmpeg_arg = None
        self.cancel = False
        self.created_time = datetime.now().strftime('%m-%d %H:%M:%S')
        self.savepath = None
        self.filename = None
        self.filepath = None
        self.quality = None
        self.headers = None
        #FfmpegQueueEntity.static_index += 1
        #FfmpegQueueEntity.entity_list.append(self)



    

    def get_video_url(self):
        return self.url

    def get_video_filepath(self):
        return self.filepath

    @abc.abstractmethod
    def refresh_status(self):
        pass

    @abc.abstractmethod
    def info_dict(self, tmp):
        pass
    
    def donwload_completed(self):
        pass

    def as_dict(self):
        tmp = {}
        tmp['entity_id'] = self.entity_id
        tmp['url'] = self.url
        tmp['ffmpeg_status'] = self.ffmpeg_status
        tmp['ffmpeg_status_kor'] = self.ffmpeg_status_kor
        tmp['ffmpeg_percent'] = self.ffmpeg_percent
        tmp['ffmpeg_arg'] = self.ffmpeg_arg
        tmp['cancel'] = self.cancel
        tmp['created_time'] = self.created_time#.strftime('%m-%d %H:%M:%S') 
        tmp['savepath'] = self.savepath
        tmp['filename'] = self.filename
        tmp['filepath'] = self.filepath
        tmp['quality'] = self.quality
        #tmp['current_speed'] = self.ffmpeg_arg['current_speed'] if self.ffmpeg_arg is not None else ''
        tmp = self.info_dict(tmp)
        return tmp


    



class FfmpegQueue(object):
    
    def __init__(self, P, max_ffmpeg_count):
        self.P = P
        self.static_index = 1
        self.entity_list = []
        self.current_ffmpeg_count = 0
        self.download_queue = None
        self.download_thread = None
        self.max_ffmpeg_count = max_ffmpeg_count
        if self.max_ffmpeg_count is None or self.max_ffmpeg_count == '':
            self.max_ffmpeg_count = 1

    def queue_start(self):
        try:
            if self.download_queue is None:
                self.download_queue = queue.Queue()
            if self.download_thread is None:
                self.download_thread = threading.Thread(target=self.download_thread_function, args=())
                self.download_thread.daemon = True  
                self.download_thread.start()
        except Exception as e: 
            self.P.logger.error(f"Exception:{str(e)}")
            self.P.logger.error(traceback.format_exc())


    def download_thread_function(self):
        while True:
            try:
                while True:
                    try:
                        if self.current_ffmpeg_count < self.max_ffmpeg_count:
                            break
                        time.sleep(5)
                    except Exception as e: 
                        self.P.logger.error(f"Exception:{str(e)}")
                        self.P.logger.error(traceback.format_exc())
                        self.P.logger.error('current_ffmpeg_count : %s', self.current_ffmpeg_count)
                        self.P.logger.error('max_ffmpeg_count : %s', self.max_ffmpeg_count)
                        break
                entity = self.download_queue.get()
                if entity.cancel:
                    continue
                
                #from .logic_ani24 import LogicAni24
                #entity.url = LogicAni24.get_video_url(entity.info['code'])
                video_url = entity.get_video_url()
                if video_url is None:
                    entity.ffmpeg_status_kor = 'URL실패'
                    entity.refresh_status()
                    #plugin.socketio_list_refresh()
                    continue

                import ffmpeg

                #max_pf_count = 0 
                #save_path = ModelSetting.get('download_path')
                #if ModelSetting.get('auto_make_folder') == 'True':
                #    program_path = os.path.join(save_path, entity.info['filename'].split('.')[0])
                #    save_path = program_path
                #try:
                #    if not os.path.exists(save_path):
                #        os.makedirs(save_path)
                #except:
                #    logger.debug('program path make fail!!')
                # 파일 존재여부 체크
                filepath = entity.get_video_filepath()
                if os.path.exists(filepath):
                    entity.ffmpeg_status_kor = '파일 있음'
                    entity.ffmpeg_percent = 100
                    entity.refresh_status()
                    #plugin.socketio_list_refresh()
                    continue
                dirname = os.path.dirname(filepath)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                f = ffmpeg.Ffmpeg(video_url, os.path.basename(filepath), plugin_id=entity.entity_id, listener=self.ffmpeg_listener, call_plugin=self.P.package_name, save_path=dirname, headers=entity.headers)
                f.start()
                self.current_ffmpeg_count += 1
                self.download_queue.task_done()    
            except Exception as e: 
                self.P.logger.error(f"Exception:{str(e)}")
                self.P.logger.error(traceback.format_exc())

    def ffmpeg_listener(self, **arg):
        import ffmpeg
        entity = self.get_entity_by_entity_id(arg['plugin_id'])
        if entity is None:
            return
        if arg['type'] == 'status_change':
            if arg['status'] == ffmpeg.Status.DOWNLOADING:
                pass
            elif arg['status'] == ffmpeg.Status.COMPLETED:
                entity.donwload_completed()
            elif arg['status'] == ffmpeg.Status.READY:
                pass
        elif arg['type'] == 'last':
            self.current_ffmpeg_count += -1
        elif arg['type'] == 'log':
            pass
        elif arg['type'] == 'normal':
            pass
        
        entity.ffmpeg_arg = arg
        entity.ffmpeg_status = int(arg['status'])
        entity.ffmpeg_status_kor = str(arg['status'])
        entity.ffmpeg_percent = arg['data']['percent']
        entity.ffmpeg_arg['status'] =  str(arg['status'])
        #self.P.logger.debug(arg)
        #import plugin
        #arg['status'] = str(arg['status'])
        #plugin.socketio_callback('status', arg)
        entity.refresh_status()


        #FfmpegQueueEntity.static_index += 1
        #FfmpegQueueEntity.entity_list.append(self)

    
    def add_queue(self, entity):
        try:
            #entity = QueueEntity.create(info)
            #if entity is not None:
            #    LogicQueue.download_queue.put(entity)
            #    return True
            entity.entity_id = self.static_index
            self.static_index += 1
            self.entity_list.append(entity)
            self.download_queue.put(entity)
            return True
        except Exception as e:
            self.P.logger.error(f"Exception:{str(e)}")
            self.P.logger.error(traceback.format_exc())
        return False

    
    def set_max_ffmpeg_count(self, max_ffmpeg_count):
        self.max_ffmpeg_count = max_ffmpeg_count

    def get_max_ffmpeg_count(self):
        return self.max_ffmpeg_count


    


    def command(self, cmd, entity_id):
        self.P.logger.debug('command :%s %s', cmd, entity_id)
        ret = {}
        try:
            if cmd == 'cancel':
                self.P.logger.debug('command :%s %s', cmd, entity_id)
                entity = self.get_entity_by_entity_id(entity_id)
                if entity is not None:
                    if entity.ffmpeg_status == -1:
                        entity.cancel = True
                        entity.ffmpeg_status_kor = "취소"
                        #entity.refresh_status()
                        ret['ret'] = 'refresh'
                    elif entity.ffmpeg_status != 5:
                        ret['ret'] = 'notify'
                        ret['log'] = '다운로드중 상태가 아닙니다.'
                    else:
                        idx = entity.ffmpeg_arg['data']['idx']
                        import ffmpeg
                        ffmpeg.Ffmpeg.stop_by_idx(idx)
                        entity.refresh_status()
                        ret['ret'] = 'refresh'
            elif cmd == 'reset':
                if self.download_queue is not None:
                    with self.download_queue.mutex:
                        self.download_queue.queue.clear()
                    for _ in self.entity_list:
                        if _.ffmpeg_status == 5:
                            import ffmpeg
                            idx = _.ffmpeg_arg['data']['idx']
                            ffmpeg.Ffmpeg.stop_by_idx(idx)
                self.entity_list = []
                ret['ret'] = 'refresh'
            elif cmd == 'delete_completed':
                new_list = []
                for _ in self.entity_list:
                    if _.ffmpeg_status_kor in [u'파일 있음', u'취소', u'사용자중지']:
                        continue
                    if _.ffmpeg_status != 7:
                        new_list.append(_)
                self.entity_list = new_list
                ret['ret'] = 'refresh'
            elif cmd == 'remove':
                new_list = []
                for _ in self.entity_list:
                    if _.entity_id == entity_id:
                        continue
                    new_list.append(_)
                self.entity_list = new_list
                ret['ret'] = 'refresh'  
            return ret
        except Exception as e:
            self.P.logger.error(f"Exception:{str(e)}")
            self.P.logger.error(traceback.format_exc())


    def get_entity_by_entity_id(self, entity_id):
        for _ in self.entity_list:
            if _.entity_id == entity_id:
                return _
        return None


    def get_entity_list(self):
        ret = []
        for x in self.entity_list:
            tmp = x.as_dict()
            ret.append(tmp)
        return ret

    


import os
import sys
import traceback

try:
    import oauth2client
except:
    os.system('pip install oauth2client')
    import oauth2client

from oauth2client import tools
from oauth2client.client import flow_from_clientsecrets
from oauth2client.file import Storage

try:
    from apiclient.discovery import build
except:
    os.system('pip install google-api-python-client')
    from apiclient.discovery import build

try:
    import time

    import gspread
    from gspread_formatting import (cellFormat, color, format_cell_range,
                                    textFormat)
except:
    os.system('pip3 install gspread')
    os.system('pip3 install gspread_formatting')
    import gspread, time
    from gspread_formatting  import cellFormat, textFormat, color, format_cell_range

from support import d, get_logger

logger = get_logger()




        
class GoogleSheetBase:
        
    current_flow = None

    color_format = {
        'green' : cellFormat(
            backgroundColor=color(0, 1, 0), #set it to yellow
            textFormat=textFormat(foregroundColor=color(0, 0, 0)),
        ),
        'yellow' : cellFormat(
            backgroundColor=color(1, 1, 0), #set it to yellow
            textFormat=textFormat(foregroundColor=color(0, 0, 0)),
        ),
        'white' : cellFormat(
            backgroundColor=color(1, 1, 1), #set it to yellow
            textFormat=textFormat(foregroundColor=color(0, 0, 0)),
        )
    }


    def __init__(self, doc_id, credentials_filepath, tab_index, unique_header):
        self.credentials_filepath = credentials_filepath
        self.credentials = self.get_credentials()
        self.doc_id = doc_id
        doc_url = f'https://docs.google.com/spreadsheets/d/{doc_id}'
        gsp = gspread.authorize(self.credentials)
        doc = gsp.open_by_url(doc_url)
        self.tab_index = tab_index
        self.ws = doc.get_worksheet(tab_index)
        self.header_info = None
        self.header_info_reverse = None
        self.unique_header = unique_header


    def get_credentials(self, project_filepath=None):
        if os.path.exists(self.credentials_filepath) == False:
            logger.info(f"credentials_filepath : {self.credentials_filepath}")
            url = self.__make_token_cli(project_filepath)
            logger.debug(f"Auth URL : {url}")
            code = input("Input Code : ")
            self.__save_token(self.credentials_filepath, code)

        store = Storage(self.credentials_filepath)
        credentials = store.get()
        if not credentials or credentials.invalid:
            logger.warning('credentials error')
            #flow = client.flow_from_clientsecrets('credentials.json', SCOPES)
            #creds = tools.run_flow(flow, store)
            os.remove(self.credentials_filepath)
            return self.get_credentials(self.credentials_filepath)
        return credentials
    
    
    def __make_token_cli(self, project_filepath):
        try:
            if project_filepath == None:
                project_filepath = os.path.join(os.path.dirname(__file__), 'cs.json')
            self.current_flow  = flow_from_clientsecrets(
                project_filepath,  # downloaded file
                'https://www.googleapis.com/auth/drive',  # scope
                redirect_uri='urn:ietf:wg:oauth:2.0:oob')
            return self.current_flow.step1_get_authorize_url()
        except Exception as e:
            logger.error(f"Exception: {e}")
            logger.error(traceback.format_exc())


    def __save_token(self, credentials_filepath, code):
        try:
            credentials = self.current_flow.step2_exchange(code)
            storage = Storage(credentials_filepath)
            storage.put(credentials)
            return True
        except Exception as e:
            logger.error(f"Exception: {e}")
            logger.error(traceback.format_exc())
            return False
    

    def get_sheet_data(self):
        tmp = self.ws.get_all_values()#[:-1]
        self.set_sheet_header(tmp[0])
        rows = tmp[1:]
        ret = []
        for row in rows:
            item = {}
            for idx, col in enumerate(row):
                item[self.header_info_reverse[idx+1]] = col
            ret.append(item)
        return ret


    def set_sheet_header(self, row):
        self.header_info = {}
        self.header_info_reverse = {}
        for idx, col in enumerate(row):
            self.header_info[col] = idx + 1
            self.header_info_reverse[idx+1] = col
        logger.debug(self.header_info)

    
    def find_row_index(self, total_data, data):
        find_row_index = -1
        #find = False
        #data['IDX'] = len(total_data)+1
        for idx, item in enumerate(total_data):
            if item[self.unique_header] == str(data[self.unique_header]):
                #find = True
                find_row_index = idx
                #data['IDX'] = find_row_index + 1
                break
        
        if find_row_index == -1:
            data['IDX'] = len(total_data)+1

        return find_row_index
    

    def sleep(self):
        time.sleep(0.5)

    def sleep_exception(self):
        time.sleep(10)

    
    def after_update_cell(self, sheet_row_index, sheet_col_index, key, value, old_value):
        pass
    

    def set_color(self, sheet_row, sheet_col1, sheet_col2, color):
        format_cell_range(self.ws, gspread.utils.rowcol_to_a1(sheet_row,sheet_col1)+':' + gspread.utils.rowcol_to_a1(sheet_row,sheet_col2), color)
    
    def set_color_row(self, sheet_row, color):
        format_cell_range(self.ws, gspread.utils.rowcol_to_a1(sheet_row,1)+':' + gspread.utils.rowcol_to_a1(sheet_row,len(self.header_info)), color)

    def set_color_cell(self, sheet_row, sheet_col, color):
        format_cell_range(self.ws, gspread.utils.rowcol_to_a1(sheet_row,sheet_col)+':' + gspread.utils.rowcol_to_a1(sheet_row,sheet_col), color)

    def write_data(self, total_data, data):
        find_row_index = self.find_row_index(total_data, data)

        write_count = 0
        for key, value in data.items():
            if key.startswith('_'):
                continue
            if value == None:
                continue
            if key not in self.header_info:
                continue
            while True:
                try:
                    if find_row_index != -1 and str(total_data[find_row_index][key]) != str(value):
                        logger.warning(f"업데이트 : {key} {total_data[find_row_index][key]} ==> {value}")
                        self.ws.update_cell(find_row_index+2, self.header_info[key], value)
                        self.after_update_cell(find_row_index+2, self.header_info[key], key, value, total_data[find_row_index][key])
                        write_count += 1
                        self.sleep()
                    elif find_row_index == -1 and value != '':
                        logger.warning(f"추가 : {key} {value}")
                        self.ws.update_cell(len(total_data)+2, self.header_info[key], value)
                        self.after_update_cell(len(total_data)+2, self.header_info[key], key, value, None)
                        write_count += 1
                        self.sleep()
                    break
                except gspread.exceptions.APIError:
                    self.sleep_exception()
                except Exception as e: 
                    logger.error(f"{key} - {value}")
                    logger.error(f"Exception:{str(e)}")
                    logger.error(traceback.format_exc())
                    logger.error(self.header_info)
                    self.sleep_exception()
          
        if find_row_index == -1:
            total_data.append(data)
        else:
            total_data[find_row_index] = data
        return write_count



import os

from support import d, logger

try:
    from selenium import webdriver
except:
    os.system("pip install --upgrade selenium")
    from selenium import webdriver

import base64
import threading
import time
import traceback

from selenium import webdriver
from selenium.common.exceptions import UnexpectedAlertPresentException
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select, WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager


class SupportSimpleSelenium(object):
    
    def __init__(self, P, mode="local", headless=False, remote=None):
        self.P = P
        self.driver = None
        self.timeout = 5
        self.driver_init(mode=mode, headless=headless, remote=remote)


    def driver_init(self, mode='local', headless=False, remote=None):
        if mode == 'local':
            from selenium.webdriver.chrome.options import Options
            options = Options()
            if headless:
                options.add_argument('headless')
            self.driver = webdriver.Chrome(service=Service(ChromeDriverManager().install()), options=options)
        elif mode == 'remote':
            from selenium.webdriver.chrome.options import Options
            options = Options()
            #options.set_preference("general.useragent.override", "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36")
            #options.set_preference("general.platform.override", "Win32")
            self.driver = webdriver.Remote(remote, options=options)


    def get_pagesoruce(self, url, wait_xpath="/html/body", retry=True):
        try:
            self.driver.get(url)
            WebDriverWait(self.driver, self.timeout).until(lambda driver: driver.find_element(By.XPATH, wait_xpath))
            return self.driver.page_source
        except Exception as e: 
            logger.error('Exception:%s', e)
            logger.error(traceback.format_exc()) 
            self.driver_quit()
            if retry:
                return self.get_pagesoruce(url, wait_xpath=wait_xpath, retry=False)


    def driver_quit(self):
        if self.driver != None:
            def func():
                self.driver.quit()
                self.driver = None
                #self.logger.debug('driver quit..')
            th = threading.Thread(target=func, args=())
            th.setDaemon(True)
            th.start()
    
    def get_downloaded_files(self):
        if not self.driver.current_url.startswith("chrome://downloads"):
            self.driver.get("chrome://downloads/")
        #driver.implicitly_wait(4)
        self.driver.implicitly_wait(2)
        return self.driver.execute_script( \
            "return  document.querySelector('downloads-manager')  "
            " .shadowRoot.querySelector('#downloadsList')         "
            " .items.filter(e => e.state === 'COMPLETE')          "
            " .map(e => e.filePath || e.file_path || e.fileUrl || e.file_url); ")
    


    def get_file_content(self, path):
        elem = self.driver.execute_script( \
            "var input = window.document.createElement('INPUT'); "
            "input.setAttribute('type', 'file'); "
            "input.hidden = true; "
            "input.onchange = function (e) { e.stopPropagation() }; "
            "return window.document.documentElement.appendChild(input); " )

        elem._execute('sendKeysToElement', {'value': [ path ], 'text': path})

        result = self.driver.execute_async_script( \
            "var input = arguments[0], callback = arguments[1]; "
            "var reader = new FileReader(); "
            "reader.onload = function (ev) { callback(reader.result) }; "
            "reader.onerror = function (ex) { callback(ex.message) }; "
            "reader.readAsDataURL(input.files[0]); "
            "input.remove(); "
            , elem)

        if not result.startswith('data:') :
            raise Exception("Failed to get file content: %s" % result)

        return base64.b64decode(result[result.find('base64,') + 7:])
    

    


    

    # docker run -d --name selenium_chromium -it -p 4446:4444 -p 5902:5900 -p 7902:7900 --shm-size 2g seleniarm/standalone-chromium:latest


import traceback

from . import logger


class ToolUtil(object):

    @classmethod
    def make_apikey_url(cls, url):
        from framework import F
        if not url.startswith('http'):
            url = F.SystemModelSetting.get('ddns') + url
        if F.SystemModelSetting.get_bool('use_apikey'):
            if url.find('?') == -1:
                url += '?'
            else:
                url += '&'
            url += f"apikey={F.SystemModelSetting.get('apikey')}"
        return url
    
    @classmethod
    def make_path(cls, data):
        from framework import F
        return data.replace('{PATH_DATA}', F.config['path_data'])
    

    @classmethod
    def run_system_command_by_id(cls, command_id):
        try:
            from system.setup import P as PP
            page_ins = PP.logic.get_module('tool').get_page('command')
            thread = page_ins.execute_thread_start(command_id)
            return thread
        except Exception as e: 
            logger.error(f'Exception:{str(e)}')
            logger.error(traceback.format_exc())  

